package net.sourceforge.ppf.plugin.bf2botplugin;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;
import java.util.Properties;
import java.util.logging.Logger;

import net.sourceforge.ppf.PPF;
import net.sourceforge.ppf.PPFPlugin;
import net.sourceforge.ppf.util.PPFHelp;

import org.xml.sax.ContentHandler;

public class BF2BotPlugin extends PPFPlugin {
    
    private static final org.apache.commons.logging.Log LOG = org.apache.commons.logging.LogFactory
            .getLog(BF2BotPlugin.class);

    private static final String PLUGIN_VERSION = "0.1";
    public static final String PLUGIN_ID = "BF2BOT";

    HashMap config = new HashMap();
    ContentHandler handler = null;
    public static String REPORT_CHANNEL = null;
    public static int CHAT_LEVEL = 0;
    public static boolean FLAG_MODE = false;

    private String commandStart = "";
    private int authLevelStart = 0;
    private int outputStart = 0; 
    private String commandStop = "";
    private int authLevelStop = 0;
    private int outputStop = 0; 
    private String commandChatLevel = "";
    private int authLevelChatLevel = 0;
    private int outputChatLevel = 0; 
    
    public static String SERVER_HOST = "";
    public static int SERVER_PORT = 0;
    private String serverPassword = "";
    
//    private String gameServerType = "";
//    private String gameServerIP = "";
//    private int gameServerPort = 0;
//    private int localPort = 0;
    
    private long showTicketDelay = 120000;

    protected static Logger logger = Logger.getLogger(BF2BotPlugin.PLUGIN_ID);

    BF2Thread bf2Thread = null;
    Properties mappings = new Properties();

    public BF2BotPlugin() {
        loadDoc("BF2BotConfig.xml");

        commandStart = getTextFromDoc("/bf2botConfig/commands/commandStart");
        authLevelStart = authLevelFromString(getAttributeFromDoc("/bf2botConfig/commands/commandStart", "authLevel"));
        outputStart = outputFromString(getAttributeFromDoc("/bf2botConfig/commands/commandStart", "output"));

        commandStop = getTextFromDoc("/bf2botConfig/commands/commandStop");
        authLevelStop = authLevelFromString(getAttributeFromDoc("/bf2botConfig/commands/commandStop", "authLevel"));
        outputStop = outputFromString(getAttributeFromDoc("/bf2botConfig/commands/commandStop", "output"));

        commandChatLevel = getTextFromDoc("/bf2botConfig/commands/commandChatLevel");
        authLevelChatLevel = authLevelFromString(getAttributeFromDoc("/bf2botConfig/commands/commandChatLevel", "authLevel"));
        outputChatLevel = outputFromString(getAttributeFromDoc("/bf2botConfig/commands/commandChatLevel", "output"));

        SERVER_HOST = getAttributeFromDoc("/bf2botConfig/server", "host");
        SERVER_PORT = Integer.valueOf(getAttributeFromDoc("/bf2botConfig/server", "port")).intValue();
        serverPassword = getAttributeFromDoc("/bf2botConfig/server", "password");
        
//        gameServerType = getAttributeFromDoc("/bf2botConfig/gameserver", "type");
//        gameServerIP = getAttributeFromDoc("/bf2botConfig/gameserver", "ip");
//        gameServerPort = Integer.valueOf(getAttributeFromDoc("/bf2botConfig/gameserver", "queryport")).intValue();
//        localPort = Integer.valueOf(getAttributeFromDoc("/bf2botConfig/gameserver", "localport")).intValue();
        
        config.put("team1Colour", "\\" + getTextFromDoc("/bf2botConfig/output/team1Colour"));
        config.put("team2Colour", "\\" + getTextFromDoc("/bf2botConfig/output/team2Colour"));

        config.put("ticketStatus", getTextFromDoc("/bf2botConfig/output/ticketStatus"));

        config.put("gameStatePlaying", getTextFromDoc("/bf2botConfig/output/gameStatePlaying"));
        config.put("gameStateEndGame", getTextFromDoc("/bf2botConfig/output/gameStateEndGame"));
        config.put("gameStatePreGame", getTextFromDoc("/bf2botConfig/output/gameStatePreGame"));
        config.put("gameStatePaused", getTextFromDoc("/bf2botConfig/output/gameStatePaused"));
        config.put("gameStateRestart", getTextFromDoc("/bf2botConfig/output/gameStateRestart"));
        config.put("gameStateNotConnected", getTextFromDoc("/bf2botConfig/output/gameStateNotConnected"));
        
        config.put("playerConnect", getTextFromDoc("/bf2botConfig/output/playerConnect"));
        config.put("playerChangeTeam", getTextFromDoc("/bf2botConfig/output/playerChangeTeam"));
        config.put("playerRevived", getTextFromDoc("/bf2botConfig/output/playerRevived"));
        config.put("playerKilledSelf", getTextFromDoc("/bf2botConfig/output/playerKilledSelf"));
        config.put("playerTeamkilled", getTextFromDoc("/bf2botConfig/output/playerTeamkilled"));
        config.put("playerKilled", getTextFromDoc("/bf2botConfig/output/playerKilled"));
        config.put("playerDeath", getTextFromDoc("/bf2botConfig/output/playerDeath"));
        config.put("playerDisconnect", getTextFromDoc("/bf2botConfig/output/playerDisconnect"));

        config.put("controlPointCapture", getTextFromDoc("/bf2botConfig/output/controlPointCapture"));
        config.put("controlPointNeutralised", getTextFromDoc("/bf2botConfig/output/controlPointNeutralised"));

        config.put("chatServer", getTextFromDoc("/bf2botConfig/output/chatServer"));
        config.put("chatPlayer", getTextFromDoc("/bf2botConfig/output/chatPlayer"));

        setVersion(PLUGIN_VERSION);

        setHelp(commandStart, new PPFHelp(commandStart, 
            "Start the BF2 match reporting", authLevelStart));
        setHelp(commandStop, new PPFHelp(commandStop, 
            "Stop the BF2 match reporting", authLevelStop));
/*        setHelp(commandChatLevel, new PPFHelp(commandChatLevel +" <chat-level>", 
            "Set which chat messages are shown for the reporter. OFF, PUBLIC" +
            ", ALL chat", authLevelChatLevel));
*/
    }

    public String getBasePath() {
        String bPath = getUrlStr();
        String retPath = "";
        if(bPath.indexOf(":") > 0) {
            retPath = bPath.substring(bPath.indexOf(":")+1); 
        }
        return retPath;
    }

    public void init(PPF ppf) {
        super.init(ppf);

        setVersion(PLUGIN_VERSION);
        
        try {
            mappings.load(new FileInputStream(getBasePath() + "bf2bot_mapping.properties"));
        } catch (FileNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }
    
    protected void onMessage(String channel, String sender, String login, String hostname,
        String message) {

        if(message.equalsIgnoreCase(commandStart)) {
            if(getBot().isAuthed(sender, authLevelStart)) {
                REPORT_CHANNEL = channel;
                bf2Thread = new BF2Thread(this);
                bf2Thread.start();
                sendOutput(outputStart, channel, sender, "Reporting started");
            }
        } else if(message.equalsIgnoreCase(commandStop)) {
            if(getBot().isAuthed(sender, authLevelStart)) {
                REPORT_CHANNEL = null;
                sendOutput(outputStop, channel, sender, "Reporting stopped");
            }
/*        } else if(message.toLowerCase().startsWith("rcon say ")) {
            if(getBot().isAuthed(sender, PPF.AUTH_TRUSTED)) {
                sendCommand("game.sayAll \""+ message.substring(9) +"\"");
            }*/
        }
    }
/*
    private void sendCommand(String command) {
        DatagramPacket packet;
        byte[] data; // For data to be Sent in packets 

        try {
            data = command.getBytes("ISO-8859-1");
        } catch(UnsupportedEncodingException e) {
            data = command.getBytes();
        }

        InetAddress inet;
        try {
            inet = InetAddress.getByName(SERVER_HOST);
            packet = new DatagramPacket(data, data.length, inet, SERVER_PORT+1);

            outSocket.send(packet);
        } catch (UnknownHostException e1) {
            // do nothing ?
            e1.printStackTrace();
        } catch (IOException e) {
            // do nothing ?
            e.printStackTrace();
        }
    }
*/
    public HashMap getConfig() {
        return config;
    }
    
    protected void unload() {
        if(bf2Thread != null) {
            bf2Thread.destroy();
        }
    }
}