package net.sourceforge.ppf;
// TODO: check plugins use of properties files
// TODO: some sort of flood prevention ... if it does flood, auto setting the message delay
// TODO: redo the properties files so that they form sentences rather than parts of sentences...
// ... this makes much more sense after translations
//TODO: automatically get a list of the available languages for the PPF part and show them with 
//the help text for !help

// TODO: 1095944834596 :irc.lut.fi 484 BotEd :Your connection is restricted! - tried to change nick

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.ConnectException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.NoRouteToHostException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLEncoder;
import java.net.UnknownHostException;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import net.sourceforge.ppf.simplewebserver.SimpleWebServer;
import net.sourceforge.ppf.util.AuthedUser;
import net.sourceforge.ppf.util.ChannelInfo;
import net.sourceforge.ppf.util.Filter;
import net.sourceforge.ppf.util.PPFHelp;
import net.sourceforge.ppf.util.PPFResourceBundle;
import net.sourceforge.ppf.util.StringUtil;

import org.apache.commons.digester.Digester;
import org.apache.commons.digester.ExtendedBaseRules;
import org.apache.commons.digester.RuleSet;
import org.apache.commons.digester.xmlrules.DigesterLoadingException;
import org.apache.commons.digester.xmlrules.FromXmlRuleSet;
import org.apache.commons.digester.xmlrules.XmlLoadException;
import org.jibble.pircbot.DccChat;
import org.jibble.pircbot.DccFileTransfer;
import org.jibble.pircbot.IrcException;
import org.jibble.pircbot.NickAlreadyInUseException;
import org.jibble.pircbot.PircBot;
import org.jibble.pircbot.User;
import org.xml.sax.SAXException;

/**
 * PPF main class, handles startup, dis/re/connecting, core bot commands, plugin
 * re/un/loading and dispatching to plugins.
 * 
 * @author DeadEd
 */
public class PPF extends PircBot {
    private static final org.apache.commons.logging.Log LOG = org.apache.commons.logging.LogFactory
            .getLog(PPF.class);
    
	private static final String HOMEPAGE = "http://ppf.sourceforge.net/";
    public static final int AUTH_ADMIN = 1;
    public static final int AUTH_MASTER = 5;
    public static final int AUTH_TRUSTED = 10;
    public static final int AUTH_ANY = AUTH_TRUSTED; // same as lowest auth level
    public static final int AUTH_NONE = 9999;
    
    private static final String DEFAULT_CONFIG_FILENAME = "PPFConfig.xml";
    private static final long DEFAULT_RETRY_DELAY = 15000L; // 15s
    private static final int DEFAULT_CONNECTION_ATTEMPTS = 20;
    private static final String DEFAULT_WEBSERVER_DOCROOT = "public_web";
    private static final int DEFAULT_WEBSERVER_PORT = 80;
    private static final String DEFAULT_PLUGINS_DIR = "..";

    private static final String VERSION_PROPERTIES = "/META-INF/version.properties";
    private static final String BOT_RULES = "digester-rules-bot.xml";
    private static final String PLUGINS_RULES = "digester-rules-plugins.xml";

    private static final int EXIT_CODE_SHUTDOWN = 0;
    private static final int EXIT_CODE_FATAL = 1;
    private static final int EXIT_CODE_RESTART = 9;
    
    private static String version;
    private static String fullVersion;
    
    static {
        InputStream is = null;
        try {
            is = PPF.class.getResourceAsStream(VERSION_PROPERTIES);
            if (is == null) {
                LOG.error("Couldn't find " + VERSION_PROPERTIES
                        + " on the classpath.");
                fullVersion = version = "PPF-custom";
            } else {
                Properties props = new Properties();
                props.load(is);
                version = props.getProperty("build.name") + " V"
                        + props.getProperty("build.version");
                fullVersion = props.getProperty("build.name") + " V"
                        + props.getProperty("build.version") + ", compiled on "
                        + props.getProperty("build.date");
            }
        } catch (IOException e) {
            LOG.error("Error reading " + VERSION_PROPERTIES
                    + " from classpath");
            fullVersion = version = "PPF-custom";
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException e1) {
                    LOG.warn("Error closing file " + VERSION_PROPERTIES);
                }
            }
        }
    }
    
	private Map authedUsers = new HashMap();
    // these are the plugins that are actually loaded
    private Map loadedPlugins = new HashMap();
    private Map currentChannels = new HashMap();
    private Date startTime;
    private Map userLangs = new HashMap();
    private PPFResourceBundle ppfResBun;
    private Map help = new HashMap();
    private int currentServer = -1;
    private boolean doQuit = false; // set to true when quitting / restarting
    private int quitExitCode; // the exit code to use when quitting

    //
    // Config
    //
    private int[] cfgDCCPorts = null;
    private boolean startWebServer = false;
    private String webServerDocumentRoot = DEFAULT_WEBSERVER_DOCROOT;
    private int webServerPort = DEFAULT_WEBSERVER_PORT;
    private String cfgFilename = DEFAULT_CONFIG_FILENAME;
    private String language = "";
    private String adminPassword = "";
    private String masterPassword = "";
    private String trustedPassword = "";
    private String encoding = "";
    private List servers = new ArrayList(4);
    private long retryDelay = DEFAULT_RETRY_DELAY;
    private int maxConnectionAttempts = DEFAULT_CONNECTION_ATTEMPTS;
	private boolean hidePingPong = false;
    private String identNick = "";
    private List nicks = new ArrayList(4);
    private List channels = new ArrayList(8);
    private boolean rejoinOnKick = false;
    private String onJoinMessage = "";
    private String commandHelp = "";
    private String commandSetLang = "";
    private String pluginsDir = DEFAULT_PLUGINS_DIR;
    // these are the plugins that are in the configuration
    private Map plugins = new HashMap();
    //topic
    private HashMap liveInfo = new HashMap();

    public PPF() {
        startTime = new Date();
        setVersion(version);
        LOG.info(version + " starting");
    }

    /**
     * Loads messages &amp; plugins, starts the webserver, connects to a IRC
     * server and joins the configured channels.
     */
    public void run() throws IOException {
        ppfResBun = new PPFResourceBundle("messages");
        Filter filter = new Filter("properties");
        // try to get the bin dir - which means we are in the development
        // environment
        File dir = new File("." + File.separator + "bin");
        if (!dir.exists()) {
            // if the bin dir doesn't exist, this is runtime
            dir = new File(".");
        }
        String[] strs = dir.list(filter);
        for (int i = 0; i < strs.length; i++) {
            int start = strs[i].lastIndexOf("_") + 1;
            int end = strs[i].indexOf(".", start);
            String langStr = strs[i].substring(start, end);
            ppfResBun.addLang(langStr);
        }

        loadPlugins();

        if (startWebServer) {
            try {
               new SimpleWebServer(new File(
                        webServerDocumentRoot), webServerPort, this);
            } catch (IOException e2) {
                LOG.error("Error starting webserver", e2);
            }
        }

        doConnect();

        joinIn();

    }

    /**
     * Load main configuration and set properties.
     * 
     * @throws DigesterLoadingException
     *             if there's a problem loading the config
     */
    public void configure(String cfgFilename, boolean pluginsOnly) {

        Digester digester = new Digester();
        digester.setRules(new ExtendedBaseRules());
        String rules = pluginsOnly ? PLUGINS_RULES : BOT_RULES;
        // try to get the bin dir - which means we are in the development environment
        File dir = new File("."+ File.separator +"bin");
        if(!dir.exists()) {
            // if the bin dir doesn't exist, this is runtime
            dir = new File(".");
        }
        try {
            URL rulesURL = new File(dir, rules).toURL();
            RuleSet ruleSet = new FromXmlRuleSet(rulesURL);
            digester.addRuleSet(ruleSet);
            digester.push(this);
        } catch (XmlLoadException e) {
            LOG.fatal("Error parsing rules", e);
            throw new RuntimeException(e);
        } catch (MalformedURLException e) {
            LOG.fatal("Invalid rules file", e);
            throw new RuntimeException(e);
        }
        
        try {
        	// TODO [xkr47] if the xml file is not valid xml, we might get strange errors
        	// from digester.parse() like "Bean has no property named servers" - maybe we
        	// should validate the xml as well to catch those ? 
            digester.parse(new File(cfgFilename));
        } catch (IOException e) {
            LOG.fatal("Error reading config file: " + cfgFilename);
            throw new RuntimeException(e);
        } catch (SAXException e) {
            LOG.fatal("Error parsing config file: " + cfgFilename);
            throw new RuntimeException(e);
        }

        String nick = "";
        String firstNick = (String) this.nicks.get(0);
        if(this.identNick != null && this.identNick.length() >= 1) {
            nick = this.identNick;
        } else {
            nick = firstNick;
        }
        setName(nick);
        setLogin(nick);

        if(identNick != null && identNick.length() >= 1) {
            setName(firstNick);
            changeNick(firstNick);
        }

        // help for core commands
        // auth lang join leave nick stats users plugins loadplugin unloadplugin reloadplugin die
        this.help.put(this.commandHelp, new PPFHelp(this.commandHelp,
                "Display the online help.", AUTH_NONE));
        this.help.put("auth", new PPFHelp("(PM) auth",
                "Authenticate as registered user", AUTH_NONE));
        this.help.put(this.commandSetLang, new PPFHelp(this.commandSetLang,
                "Set your preferred language", AUTH_NONE));
        this.help.put("join", new PPFHelp("(PM) join <channel>",
                "Make the bot join a channel", AUTH_ADMIN));
        this.help.put("leave", new PPFHelp("(PM) leave <channel>",
                "Make the bot leave a channel", AUTH_ADMIN));
        this.help.put("nick", new PPFHelp("(PM) nick <name>",
                "Change the bot's nick", AUTH_ADMIN));
        this.help.put("stats", new PPFHelp("(PM) stats",
                "Display basic system stats", AUTH_ADMIN));
        this.help.put("users", new PPFHelp("(PM) users",
                "Display currently authenticated users", AUTH_ADMIN));
        this.help.put("plugins", new PPFHelp("(PM) plugins",
                "Display configured plugins", AUTH_ADMIN));
        this.help.put("loadplugin", new PPFHelp("(PM) loadplugin <plugin>",
                "Load a plugin", AUTH_ADMIN));
        this.help.put("unloadplugin", new PPFHelp("(PM) unloadplugin <plugin>",
                "Unload a plugin", AUTH_ADMIN));
        this.help.put("reloadplugin", new PPFHelp("(PM) reloadplugin <plugin>",
                "Reload a plugin", AUTH_ADMIN));
        this.help.put("die", new PPFHelp("(PM) die",
                "Shutdown the bot", AUTH_ADMIN));
        
        this.cfgFilename = cfgFilename;
    }

    /**
     * Connect to the configured servers. If there's a problem connecting,
     * try the next server. If the configured nick is already in use, try the
     * next nick. This method blocks until the bot is successfully connected.
     */
    private void doConnect() {
        String initialNick = getName();
        IrcServer server = getNextServer();
        boolean tryNextNick = false;
        boolean connected = false;
        int connectionAttempts = 0;
        int iterations = 0;

        while(!connected && iterations < this.maxConnectionAttempts) {
            // see if we need to try another nick, and maybe sleep for a
            // while if we tried all
            if(tryNextNick) {
                String nextNick = getNextNick(getName());
                if(nextNick.equals(initialNick)) {
                    LOG.warn(ppfResBun.getString(language, "all_nicks_used",
                        "All given nicks are in use"));
                    // sleep for a while, we've tried all nicks during this doConnect call

                    LOG.warn("Waiting for " + (getRetryDelay() / 1000L)
                            + " seconds before retrying...");
                    try {
                        Thread.sleep(getRetryDelay());
                    } catch (InterruptedException e) {
                        // doesn't matter
                    }
                }
                setName(nextNick);
                setLogin(nextNick);
              }

              // try the next server, and maybe sleep for a while if we tried all
              if (connectionAttempts > 0) {
                  server = getNextServer();
                  if(this.currentServer == 0) {
                      iterations++;
                      // sleep for a while, we've tried all servers during this doConnect() call
                      LOG.warn("Could not connect to any of the configured servers.");
                      if (getRetryDelay() > 0) {
                          LOG.warn("Waiting for " + (getRetryDelay() / 1000L)
                                  + " seconds before retrying...");
                          try {
                              Thread.sleep(getRetryDelay());
                          } catch (InterruptedException e) {
                              // doesn't matter
                          }
                      }
                  } 
              }

              if (LOG.isInfoEnabled()) {
                  LOG.info("[" + getName() + "] Connecting to "
                        + server.getName() + ":" + server.getPort()
                        + ": (" + (connectionAttempts + 1)
                        + "/" + this.maxConnectionAttempts + ")");
              }
              connectionAttempts++;

              try {
                  connect(server.getName(), server.getPort(), server.getPassword());
                  // if we get here, without exceptions, all is fine.
                  connected = true;
                  if (LOG.isInfoEnabled()) {
                      LOG.info("[" + getName() + "] Connected to "
                            + server.getName() + ":" + server.getPort());
                }
        
              } catch (NickAlreadyInUseException naiuex) {
                  tryNextNick = true;
                  LOG.warn("Nickname in use, trying next nick", naiuex);
              } catch (IrcException iex) {
                  if(iex.getMessage().indexOf(" 437 ") != -1) {
                      // 437 means nick/channel temporarily unavailable. in this
                      // case, we haven't tried to join any channel yet, so it
                      // must be the nick.
                      tryNextNick = true;
                      LOG.warn("Nickname temporarily unavailable, trying next nick",
                              iex);
                  } else {
                      LOG.warn("Problem connecting. Server " + server.getName()
                            + ":" + server.getPort()
                              + " sent an error response", iex);
                  }
              } catch (ConnectException ex) {
                  LOG.warn("Problem connecting to " + server.getName() + ":"
                        + server.getPort() + ": " + ex.getLocalizedMessage());
              } catch (UnknownHostException ex) {
                  LOG.warn("Problem connecting. Unknown host: "
                        + server.getName() + ":" + server.getPort()
                        + ": " + ex.getLocalizedMessage());
              } catch (NoRouteToHostException ex) {
                  LOG.warn("Problem connecting. No route to host: "
                        + server.getName() + ":" + server.getPort()
                        + ": " + ex.getLocalizedMessage());
              } catch (IOException ioex) {
                  LOG.warn("Problem connecting to " + server.getName() + ":"
                        + server.getPort(), ioex);
              }
         }
         if (connectionAttempts >= this.maxConnectionAttempts) {
             throw new RuntimeException("Giving up after " + connectionAttempts
                    + " connection attempts");
         }
    }

    /**
     * Join all configured channels.
     */
    private void joinIn() {
        // iterate over List
        // TODO: bot still sends notice that channel is joined when a channel that is
        // passworded is given, but not password is used for the join
        for( Iterator iter = channels.iterator(); iter.hasNext(); ) {
            IrcChannel channel = (IrcChannel) iter.next();

			if(channel.getPassword() != null && channel.getPassword().length() > 0) {
				joinChannel(StringUtil.prependHash(channel.getName()), channel.getPassword());
                // add configured channels to current list
                currentChannels.put(channel.getName(), channel.getPassword());
                // TODO move putting the channel in currentChannels to onJoin()?
			} else {
				joinChannel(StringUtil.prependHash(channel.getName()));
                // add configured channels to current list
                currentChannels.put(channel.getName(), "");
                // TODO move putting the channel in currentChannels to onJoin()?
			}
        }
	}

    /**
     * This one logs to the configured log file and additionally 
     * to stdout if verbose mode is on.
     * <p>
     * <table border="1" cellspacing="0" cellpadding="6">
     * <caption>Log levels used</caption>
     * <tr>
     *     <th>Type of message</th>
     *     <th>log level</th>
     * </tr><tr>
     *     <td>Incoming line</td>
     *     <td>OFF</td>
     * </tr><tr>
     *     <td>Outgoing line</td>
     *     <td>OFF</td>
     * </tr><tr>
     *     <td>Status message</td>
     *     <td>DEBUG</td>
     * </tr><tr>
     *     <td>Errors</td>
     *     <td>ERROR</td>
     * </tr>
     * </table>
     * 
     * <p>
     * 
     * @see org.jibble.pircbot.PircBot#log(java.lang.String)
     * @since 0.5
     */
    public void log(String line) {
        //TODO this would be a good point for channel activity logging

    	// only call super.log() if it's not an unwanted PING or PONG
        if (!hidePingPong || !(line.startsWith("PING ") || line.startsWith(">>>PONG "))) {
        	super.log(line);
        }
        // log errors
        if (line.startsWith("###")) {
            LOG.error(line);
        } 
        // debug status messages
        else if (line.startsWith("***")) {
            if (LOG.isDebugEnabled()) {
                LOG.debug(line);
            }
        } 
    }
    
    protected void onJoin(String channel, String sender, String login, String hostname) {
        if(!sender.equals(getNick()) && onJoinMessage.length() > 0) {
    		sendNotice(sender, onJoinMessage);
        }

        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onJoin(channel, sender, login, hostname);
            } catch (Exception e) {
                LOG.error("Error handling join in plugin " + plugin.getClass().getName(), e);
            }
        }

        if (sender.equals(getNick())) {
            if (LOG.isInfoEnabled()) {
                // TODO i18n
                LOG.info("[" + getName() + "] Joined channel: " + channel);
            }
        }
	}

    protected void checkPublicCommand(String sender,String message) {

        // TODO: externalise the strings
        if(message.trim().equalsIgnoreCase(commandHelp)) {
        	StringBuffer helpCmds = new StringBuffer("Bot commands are: auth "+ commandSetLang);
        	if(isAuthed(sender, AUTH_ADMIN)) {
				helpCmds.append(" join leave nick stats users plugins loadplugin unloadplugin reloadplugin die "); 
        	}
            sendNotice(sender, helpCmds.toString());
            StringBuffer plugList = new StringBuffer("Plugins available:");
            
            for( Iterator iter = plugins.values().iterator(); iter.hasNext(); ) {
                PPFPluginInfo pi = (PPFPluginInfo) iter.next();
                if(pi.isLoaded()) {
                    PPFPlugin pluginInstance = (PPFPlugin) loadedPlugins.get(pi.getName());

					HashMap plugHelp = pluginInstance.getHelp();
					Iterator plugHelpIter = plugHelp.values().iterator();
					while(plugHelpIter.hasNext()) {
						PPFHelp helpObj = (PPFHelp) plugHelpIter.next();
						if(isAuthed(sender, helpObj.getAuthLevel())) {
							// if there is even one command allowed, then show the plugin on the list
							plugList.append(" "+ pi.getName());
							break;
						}
					}
                }
            }
            sendNotice(sender, plugList.toString());
            sendNotice(sender, "For more help, type: " + this.commandHelp
                    + " <plugin-name> or " + this.commandHelp + " <command>");
              
        } else if(message.trim().startsWith(commandHelp + " ")) {
            String helpTopic = message.substring(commandHelp.length()).trim();
            boolean isPluginHelp = false;
            Iterator ksIter = loadedPlugins.keySet().iterator();
            String key = "";
            while(ksIter.hasNext()) {
                key = (String) ksIter.next();
                if(key.equalsIgnoreCase(helpTopic)) {
                    isPluginHelp = true;
                    PPFPlugin pluginInstance = (PPFPlugin) loadedPlugins.get(key);
					HashMap plugHelp = pluginInstance.getHelp();
					Iterator plugHelpIter = plugHelp.values().iterator();
					while(plugHelpIter.hasNext()) {
						PPFHelp helpObj = (PPFHelp) plugHelpIter.next();
						if(isAuthed(sender, helpObj.getAuthLevel())) {
							sendNotice(sender, helpObj.getCommand() +" - "+ helpObj.getText());
						}
					}
					break;
                }
            }
            if (!isPluginHelp) {
                PPFHelp helpObj = (PPFHelp) this.help.get(helpTopic);
                if (helpObj != null) {
                    if (isAuthed(sender, helpObj.getAuthLevel())) {
                        sendNotice(sender, helpObj.getCommand() + " - "
                                + helpObj.getText());
                    }
                }
            }

        } else if(message.trim().startsWith(commandSetLang + " ")) {
            String langWanted = message.substring(commandSetLang.length()).trim();
            // TODO: this assumes everything exists and everything is done properly
    
            userLangs.put(sender, langWanted);
            sendNotice(sender, "Language set to: " + langWanted);
        }
    }
    
    protected void onMessage(String channel, String sender, String login,
		String hostname, String message) {

        // Check for !help or !setlang
        checkPublicCommand(sender,message);

        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onMessage(channel, sender, login, hostname, message);
            } catch (Exception e) {
                LOG.error("Error handling message in plugin " + plugin.getClass().getName(), e);
            }
        }
	}

    public void fromWeb(String sender, String hostname, HashMap parameters) {
        String action = (String) parameters.get("action");
        if(action != null) {
            String message = (String) parameters.get("message");

            if(action.equalsIgnoreCase("privatemessage")) {
                onPrivateMessage(sender, "", hostname, message);
            } else if(action.equalsIgnoreCase("publicmessage")) {
                onMessage("", sender, "", hostname, message);
            }
        }
    }

    public boolean verifyAuthentication(String sender, String checkPassword, String hostname) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("hostname in verifyAuth = " + hostname);
        }
        // add the auth level depending on the password given
        if (checkPassword.equals(adminPassword)) {
			if (!isAuthed(sender, AUTH_ADMIN)) {     
                if(isAuthed(sender, AUTH_ANY)) {
                    removeAuthed(sender);               
                }
				authedUsers.put(sender, new AuthedUser(sender, hostname, AUTH_ADMIN));
			}
        } else if (checkPassword.equals(masterPassword)) {
            if (!isAuthed(sender, AUTH_MASTER)) {
                if(isAuthed(sender, AUTH_ANY)) {
                    removeAuthed(sender);               
                }
                authedUsers.put(sender, new AuthedUser(sender, hostname, AUTH_MASTER));
            }
        } else if (checkPassword.equals(trustedPassword)) {
            if (!isAuthed(sender, AUTH_TRUSTED)) {                    
                authedUsers.put(sender, new AuthedUser(sender, hostname, AUTH_TRUSTED));
            }
        } else {
        	return false;
        }
        return true;
    }
    
    protected void onPrivateMessage(String sender, String login, String hostname,
		String message) {
			
        // Check for !help or !setlang
        checkPublicCommand(sender,message);
        
		String realMsg = "";
		if (message.toUpperCase().startsWith("AUTH")) {
			String checkPassword = message.substring(4).trim();
            // add the auth level depending on the password given
            if (checkPassword.equals(adminPassword)) {
				if (!isAuthed(sender, AUTH_ADMIN)) {     
                    if(isAuthed(sender, AUTH_ANY)) {
                        removeAuthed(sender);               
                    }
					authedUsers.put(sender, new AuthedUser(sender, hostname, AUTH_ADMIN));
                    Iterator chans = getCommonChannels(sender);
                    while(chans.hasNext()) {
                        String channel = (String) chans.next();
                        onAuth(login, channel);
                    }
					sendNotice(sender, ppfResBun.getString(getUserLang(sender), "authed", "You have been AUTHed"));
                    if (LOG.isInfoEnabled()) {
                        LOG.info(StringUtil.formatUser(sender, login, hostname)
                                + " authed as ADMIN");
                    }
				} else {
					sendNotice(sender, ppfResBun.getString(getUserLang(sender), "already_authed", "You are already AUTHed"));
				}
            } else if (checkPassword.equals(masterPassword)) {
                if (!isAuthed(sender, AUTH_MASTER)) {
                    if(isAuthed(sender, AUTH_ANY)) {
                        removeAuthed(sender);               
                    }
                    authedUsers.put(sender, new AuthedUser(sender, hostname, AUTH_MASTER));
                    Iterator chans = getCommonChannels(sender);
                    while(chans.hasNext()) {
                        String channel = (String) chans.next();
                        onAuth(login, channel);
                    }
                    sendNotice(sender, ppfResBun.getString(getUserLang(sender), "authed", "You have been AUTHed"));
                    if (LOG.isInfoEnabled()) {
                        LOG.info(StringUtil.formatUser(sender, login, hostname)
                                + " authed as MASTER");
                    }
                } else {
                    sendNotice(sender, ppfResBun.getString(getUserLang(sender), "already_authed", "You are already AUTHed"));
                }
            } else if (checkPassword.equals(trustedPassword)) {
                if (!isAuthed(sender, AUTH_TRUSTED)) {                    
                    Iterator chans = getCommonChannels(sender);
                    while(chans.hasNext()) {
                        String channel = (String) chans.next();
                        onAuth(login, channel);
                    }
                    authedUsers.put(sender, new AuthedUser(sender, hostname, AUTH_TRUSTED));
                    sendNotice(sender, ppfResBun.getString(getUserLang(sender), "authed", "You have been AUTHed"));
                    if (LOG.isInfoEnabled()) {
                        LOG.info(StringUtil.formatUser(sender, login, hostname)
                                + " authed as TRUSTED");
                    }
                } else {
                    sendNotice(sender, ppfResBun.getString(getUserLang(sender), "already_authed", "You are already AUTHed"));
                }
            }
		} else if (isAuthed(sender, AUTH_ADMIN)) {
			realMsg = message;
        
			//
			// join a channel
			//		
			if (realMsg.toUpperCase().startsWith("JOIN")) {
	            String joinStr = StringUtil.prependHash(realMsg.substring(4).trim());

                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + ": join " + joinStr);
                }

				joinChannel(joinStr);
	            if(joinStr.indexOf(" ") > 0) {
	                String channel = joinStr.substring(0, joinStr.indexOf(" ")).trim();
	                String password = joinStr.substring(channel.length()).trim();
	                currentChannels.put(channel, password);
	            } else {
	                currentChannels.put(StringUtil.prependHash(realMsg.substring(4).trim()), "");
	            }
			// 
			// leave/part a channel
			//
			} else if (realMsg.toUpperCase().startsWith("LEAVE")) {
	            String leaveStr = StringUtil.prependHash(realMsg.substring(5).trim()); 

                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + ": leave " + leaveStr);
                }

				partChannel(leaveStr);
	            if(currentChannels.containsKey(leaveStr)) {
	                currentChannels.remove(leaveStr);
	            }
			} else if (realMsg.toUpperCase().startsWith("PART")) {
				String leaveStr = StringUtil.prependHash(realMsg.substring(4).trim()); 

                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + ": part " + leaveStr);
                }

                partChannel(leaveStr);
				if(currentChannels.containsKey(leaveStr)) {
					currentChannels.remove(leaveStr);
				}
			//
			// die - quit JVM
			//
			} else if (realMsg.equalsIgnoreCase("DIE")) {
                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + " requests shutdown");
                }
				quit();
            //
            // restart - quit JVM and restart
            //
            } else if (realMsg.equalsIgnoreCase("RESTART")) {
                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + " requests shutdown for a restart");
                }
                restart();
			//
			// change nick
			//
			} else if (realMsg.toUpperCase().startsWith("NICK")) {
                String oldNick = getNick();
                String newNick = realMsg.substring("NICK".length()).trim();
                
                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + " changes the bot's nick from " + oldNick
                            + " to " + newNick);
                }

                setName(newNick);
				changeNick(newNick);
	        //
	        // stats
	        //
	        } else if (realMsg.toUpperCase().startsWith("STATS")) {
                Locale l = new Locale(getUserLang(sender));
                DateFormat df = DateFormat.getDateTimeInstance(
                        DateFormat.MEDIUM, DateFormat.MEDIUM, l);
                long mem = Runtime.getRuntime().totalMemory()
                        - Runtime.getRuntime().freeMemory();
	            sendMessage(sender, ppfResBun.getString(getUserLang(sender), "stats_1", 
	                "I have been running since:") +" "+ df.format(startTime) +" "+ 
	                ppfResBun.getString(getUserLang(sender), "stats_2", "and I have used") +" "+ 
	                mem +" "+ ppfResBun.getString(getUserLang(sender), 
	                "stats_3", "bytes of memory"));
			//
			// List plugins
			//
			} else if (realMsg.toUpperCase().startsWith("PLUGINS")) {
                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + " requests plugin list");
                }

                configure(cfgFilename, true);
				
				StringBuffer plugList = new StringBuffer();
	            StringBuffer notLoaded = new StringBuffer();
	
		        for( Iterator iter = plugins.values().iterator(); iter.hasNext(); ) {
	                PPFPluginInfo pi = (PPFPluginInfo) iter.next();
	                if(pi.isLoaded()) {
	                    PPFPlugin pluginInstance = (PPFPlugin) loadedPlugins.get(pi.getName());
	                    plugList.append(pi.getName() +" (v"+ pluginInstance.getVersion() +") ");                    
	                } else {
	                    notLoaded.append(pi.getName() +" ");
	                }
		        }
	        	sendMessage(sender, ppfResBun.getString(getUserLang(sender), "loaded_plugins", 
	                "Loaded plugins are:") +" "+ plugList);
	
	            if(notLoaded.length() > 0) {
	                sendMessage(sender, ppfResBun.getString(getUserLang(sender), "not_loaded_plugins",
	                    "Plugins not loaded are:") +" "+ notLoaded);
	            }
	
			//
			// reload a plugin
			//
			} else if (realMsg.toUpperCase().startsWith("RELOADPLUGIN")) {
				String pluginName = realMsg.substring("RELOADPLUGIN".length()).trim();
                
                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + " reloads plugin " + pluginName);
                }
	
	            boolean pluginFound = false;
	            // need to go through list due to wanting a case-insensitive search
	            // TODO: make internal entries as one case now that we are using PPFPluginInfo 
	            Iterator ksIter = plugins.keySet().iterator();
	            String key = "";
	            while(ksIter.hasNext()) {
	                key = (String) ksIter.next();
	                if(key.equalsIgnoreCase(pluginName)) {
	                    pluginFound = true;
	                    break;
	                }
	            }
	              
	            if(pluginFound) {
	                if (reloadPlugin(key)) {
                        sendNotice(sender, ppfResBun.getString(
                                getUserLang(sender), "reloaded_plugin",
                                "Reloaded plugin:")
                                + " " + pluginName);
                    } else {
                        sendNotice(sender, ppfResBun.getString(
                                getUserLang(sender), "error_loading_plugin",
                                "Error loading plugin:")
                                + " " + pluginName);
                    }
	            } else {
	                sendNotice(sender, ppfResBun.getString(getUserLang(sender),
                            "plugin_not_found", "Couldn't find plugin:")
                            + " " + pluginName);
	            }
	
	        //
	        // load a plugin
	        //
	        } else if (realMsg.toUpperCase().startsWith("LOADPLUGIN")) {
	            String pluginName = realMsg.substring("LOADPLUGIN".length()).trim();
	
                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + " loads plugin " + pluginName);
                }

                if(loadedPlugins.containsKey(pluginName)) {
	                sendNotice(sender, ppfResBun.getString(getUserLang(sender), "plugin_already_loaded ",
	                    "Plugin is already loaded:") +" "+ pluginName);                
	            } else {
	                boolean pluginFound = false;
	
                    configure(cfgFilename, true);
					
	                // need to go through list due to wanting a case-insensitive search
	                // TODO: make internal entries as one case now that we are using PPFPluginInfo 
	                Iterator ksIter = plugins.keySet().iterator();
	                String key = "";
	                while(ksIter.hasNext()) {
	                    key = (String) ksIter.next();
	                    if(key.equalsIgnoreCase(pluginName)) {
	                        pluginFound = true;
	                        break;
	                    }
	                }
	                  
	                if(pluginFound) {
                        if (loadPlugin(key)) {
    						sendNotice(sender, ppfResBun.getString(getUserLang(sender), "loaded_plugin",
    							"Loaded plugin:") +" "+ pluginName);
                        } else {
                            sendNotice(sender, ppfResBun.getString(
                                    getUserLang(sender), "error_loading_plugin",
                                    "Error loading plugin:") + " " + pluginName);
                        }
	                } else {
	                    sendNotice(sender, ppfResBun.getString(getUserLang(sender), "plugin_not_found",
	                        "Couldn't find plugin:") +" "+ pluginName);
	                }
	            }
	
	        //
	        // unload a plugin
	        //
	        } else if (realMsg.toUpperCase().startsWith("UNLOADPLUGIN")) {
	            String pluginName = realMsg.substring("UNLOADPLUGIN".length()).trim();
	
                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + " unloads plugin " + pluginName);
                }

                boolean pluginFound = false;
	            Iterator ksIter = loadedPlugins.keySet().iterator();
	            String key = "";
	            while(ksIter.hasNext()) {
	                key = (String) ksIter.next();
	                if(key.equalsIgnoreCase(pluginName)) {
	                    pluginFound = true;
	                    break;
	                }
	            }
	
	            if(pluginFound) {
	                unloadPlugin(key);
	                sendNotice(sender, ppfResBun.getString(getUserLang(sender), "unloaded_plugin",
	                    "Unloaded Plugin:") +" "+ pluginName);
	            } else {
	                sendNotice(sender, ppfResBun.getString(getUserLang(sender), "plugin_not_found",
	                    "Couldn't find plugin:") +" "+ pluginName);
	            }
	
	        //
	        // check if an update is available for a plugin
	        //
	        } else if (realMsg.toUpperCase().startsWith("CHECKPLUGIN")) {
	            String pluginName = realMsg.substring("CHECKPLUGIN".length()).trim();

                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + " updates plugin " + pluginName);
                }

	// TODO: say if plugin is not loaded
	// TODO: say if plugin is not found
	            Iterator loadedIter = loadedPlugins.entrySet().iterator();
	            String key = "";
	            while(loadedIter.hasNext()) {
	                Map.Entry e = (Map.Entry) loadedIter.next();
	                key = (String) e.getKey(); 
	                if(key.equalsIgnoreCase(pluginName)) {
	                    PPFPlugin pluginInstance = (PPFPlugin) e.getValue();                    
	                    String newVersion = checkPluginUpdates(key, pluginInstance.getVersion());
	                    sendMessage(sender, "Version available for the "+ key +" plugin is: "+ newVersion);                    
	                }
	            }        
	        
	        //
	        // update the plugin (download from server)
	        //
	        } else if (realMsg.toUpperCase().startsWith("UPDATEPLUGIN")) {
	            String pluginName = realMsg.substring("UPDATEPLUGIN".length()).trim();
	    //TODO: say if plugin is not loaded
	    //TODO: say if plugin is not found
	            Iterator loadedIter = loadedPlugins.entrySet().iterator();
	            String key = "";
	            while(loadedIter.hasNext()) {
	                Map.Entry e = (Map.Entry) loadedIter.next();
	                key = (String) e.getKey(); 
	                if(key.equalsIgnoreCase(pluginName)) {
	                    PPFPlugin pluginInstance = (PPFPlugin) e.getValue();
	
	                    URL urli = pluginInstance.getClass().getProtectionDomain().getCodeSource().getLocation();
	                    String urlStr = urli.getFile();
	                    if(!urlStr.endsWith("/")) {
	                        urlStr = urlStr.substring(0, urlStr.lastIndexOf("/")+1);
	                    }
	
	                    downloadPluginUpdate(key, urlStr);                    
	                    sendMessage(sender, "Plugin "+ key +" downloaded");                    
	                    if (reloadPlugin(key)) {
                            sendNotice(sender, ppfResBun.getString(
                                    getUserLang(sender), "reloaded_plugin",
                                    "Reloaded plugin:")
                                    + " " + pluginName);
                        } else {
                            sendNotice(sender, ppfResBun.getString(
                                    getUserLang(sender), "error_loading_plugin",
                                    "Error loading plugin:")
                                    + " " + pluginName);
                        }
	                }
	            }        
	        
	        
	        //
	        // show authed users
	        //
	        } else if (realMsg.equalsIgnoreCase("USERS")) {
                if (LOG.isInfoEnabled()) {
                    LOG.info(StringUtil.formatUser(sender, login, hostname)
                            + " requests user list");
                }
	            StringBuffer authedUserStr = new StringBuffer(authedUsers.size() +" users authed");
	            
	            Iterator iter = authedUsers.values().iterator();
	            while(iter.hasNext()) {
	                AuthedUser au = (AuthedUser) iter.next();
	                authedUserStr.append(": "+ au.getNick());
	                int level = au.getAuthLevel();
	                if(level == PPF.AUTH_ADMIN) {
	                    authedUserStr.append(" [ADMIN]");
	                } 
	                if(level == PPF.AUTH_MASTER) {
	                    authedUserStr.append(" [MASTER]");
	                } 
	                if(level == PPF.AUTH_TRUSTED) {
	                    authedUserStr.append(" [TRUSTED]");
	                }
	            }
	
	            sendMessage(sender, authedUserStr.toString());	            
			} else {
				onPrivateMessageToPlugins(sender, login, hostname, message);
			}
		} else {
			onPrivateMessageToPlugins(sender, login, hostname, message);
		}
    }
	
    /**
     * Dispatch private messages to plugins, see
     * {@link onPrivateMessage(java.lang.String,java.lang.String,java.lang.String,java.lang.String)}
     */
    private void onPrivateMessageToPlugins(String sender, String login, String hostname,
   		String message) {

		Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onPrivateMessage(sender, login, hostname, message);
            } catch (Exception e) {
                LOG.error("Error handling private message in plugin " + plugin.getClass().getName(), e);
            }
        }
	}

    /**
     * Dispatch "auth" events to the plugins. Called for once for every channel
     * the authed user is on. 
     * @param login the login part of the user's hostmask
     * @param channel the channel 
     */
    protected void onAuth(String login, String channel) {
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onAuth(login, channel);
            } catch (Exception e) {
                LOG.error("Error handling auth in plugin " + plugin.getClass().getName(), e);
            }
        }
    }

    protected void onQuit(String sourceNick,  String sourceLogin,	
        String sourceHostname,	String reason) {
            
		if (isAuthed(sourceNick, AUTH_ANY) && !inCommonChannel(sourceNick)) {
			removeAuthed(sourceNick);
        }

        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onQuit(sourceNick, sourceLogin, sourceHostname, reason);
            } catch (Exception e) {
                LOG.error("Error handling quit in plugin " + plugin.getClass().getName(), e);
            }
        }
        
	}

    protected void onDisconnect() {
        if (LOG.isInfoEnabled()) {
            LOG.info("[" + getName() + "] Disconnected from server");
        }
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onDisconnect();
            } catch (Exception e) {
                LOG.error("Error handling disconnect in plugin " + plugin.getClass().getName(), e);
            }
        }

        // if the user called quit() or restart(), doQuit is true and thus we quit when disconnection succeeds
        if (doQuit) {
        	try {
        		// need to make copy of the list of loaded plugins, otherwise we get ConcurrentModificationException
	        	Set pluginsToUnload = new HashSet(loadedPlugins.keySet());
	            Iterator ksIter = pluginsToUnload.iterator();
	            while(ksIter.hasNext()) {
	                String pluginName = (String) ksIter.next();
	                unloadPlugin(pluginName);
	            }
        	} catch(Throwable e) {
        		LOG.error("Error unloading plugins", e);
        	}

            LOG.info("Exiting with code " + quitExitCode);
        	System.exit(quitExitCode);

        } else {
        	// we are not quitting, so try to reconnect
        	try {
        		doConnect();
        		joinIn();
        	} catch (RuntimeException e) {
        		LOG.fatal("Error while reconnecting, exiting now", e);
        		System.exit(EXIT_CODE_FATAL);
        	}
        }
	}

    protected void onKick(String channel, String kickerNick, String kickerLogin,
		String kickerHostname, String recipientNick, String reason) {
            
		if (isAuthed(recipientNick, AUTH_ANY) && !inCommonChannel(recipientNick)) {
			removeAuthed(recipientNick);
		}
		if (rejoinOnKick && recipientNick.equalsIgnoreCase(getName())) {

            if(currentChannels.containsKey(channel)) {
                joinChannel(channel, (String) currentChannels.get(channel));
            } else {
			    joinChannel(channel);
            }
            
			if (reason.equalsIgnoreCase(kickerNick)
				|| reason.equalsIgnoreCase(recipientNick)) {
				sendMessage(
					channel,
					kickerNick + " " + ppfResBun.getString(getUserLang(kickerNick), "not_nice",
                        "... that was not very nice!"));
			} else {
				sendMessage(channel, kickerNick	+", \""+ reason	+"\" "
                    + ppfResBun.getString(getUserLang(kickerNick), "reason_for_kick_not_good",
                    "was not a good enough reason to kick me ... be nice"));
			}
		}

        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onKick(channel, kickerNick, kickerLogin, kickerHostname,
                        recipientNick, reason);
            } catch (Exception e) {
                LOG.error("Error handling kick in plugin " + plugin.getClass().getName(), e);
            }
        }
	}

    protected void onSetChannelBan(String channel, String sourceNick,
		String sourceLogin, String sourceHostname, String hostmask) {
			
		if (isAuthed(sourceNick, AUTH_ANY) && !inCommonChannel(sourceNick)) {
			removeAuthed(sourceNick);
		}

        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onSetChannelBan(channel, sourceNick, sourceLogin,
                        sourceHostname, hostmask);
            } catch (Exception e) {
                LOG.error("Error handling channel ban in plugin " + plugin.getClass().getName(), e);
            }
        }
	}

    protected void onPart(String channel, String sender,
		String login, String hostname) {
			
		if (isAuthed(sender, AUTH_ANY) && !inCommonChannel(sender)) {
			removeAuthed(sender);
		}

        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onPart(channel, sender, login, hostname);
            } catch (Exception e) {
                LOG.error("Error handling part in plugin " + plugin.getClass().getName(), e);
            }
        }
        
        if (sender.equals(getNick())) {
            if (LOG.isInfoEnabled()) {
                //TODO i18n
                LOG.info("[" + getName() + "] Left channel: " + channel);
            }
        }
	}

    /**
     * Test if the user is authed with at least the specified auth level.
     * @param nick nick of the user who's auth status is checked.
     * @param authLevel the min. auth level required.
     * @return true if the user is authed with at least the specified authLevel.
     */
	public boolean isAuthed(String nick, int authLevel) {

		// do a quick check for NONE auth level - all users have NONE
		if(authLevel == AUTH_NONE) {
			return true;
			
		// do a quick check for ANY auth level
		} else if(authLevel == AUTH_ANY) {
            if(authedUsers.containsKey(nick)) {
                return true;
            } else {
                return false;
            }
        } else {
            // longer check for a specific auth level
            Iterator iter = authedUsers.values().iterator();
            while(iter.hasNext()) {
                AuthedUser au = (AuthedUser) iter.next();
                if (au.getNick().equals(nick)) {
                    if(au.getAuthLevel() <= authLevel) {
                        return true;
                    }
                }                
            }
        
		}

		return false;
	}

    /**
     * Called when an authed user quits, parts, gets kicked or banned or auth's
     * with a higher auth level.
     * @param nick nick of the user who's auth status changes.
     */
	private void removeAuthed(String nick) {
        if(authedUsers.containsKey(nick)) {
            authedUsers.remove(nick);
        }
	}


    /**
     * Tests if the user is in one of the channels the bot is in as well.
     * @param nick the user's nick.
     * @return true if the user is in one of the bot's channels.
     */
    public boolean inCommonChannel(String nick) {
        Iterator chans = currentChannels.keySet().iterator();
        while(chans.hasNext()) {
            String channel = (String) chans.next();
            User[] users = getUsers(channel);
            for(int x=0; x < users.length; x++) {
                User user = users[x];
                if(user.getNick().equals(nick)) {
                    return true;
                }
            }            
        }
        return false;
    }

    /**
     * Get an Iterator for all the channels the user AND the bot are in.
     * @param nick the user's nick.
     * @return Iterator for the channels the user and the bot have in common.
     */
    public Iterator getCommonChannels(String nick) {
        HashMap commonChans = new HashMap();
        
        Iterator chans = currentChannels.keySet().iterator();
        while(chans.hasNext()) {
            String channel = (String) chans.next();
            User[] users = getUsers(channel);
            for(int x=0; x < users.length; x++) {
                User user = users[x];
                if(user.getNick().equals(nick)) {
                    commonChans.put(channel, channel);
                }
            }            
        }
        
        return commonChans.keySet().iterator();
    }

    /**
     * Exits the JVM.
     */
	public void quit() {
        if (LOG.isInfoEnabled()) {
            LOG.info(version + " shutdown");
        }
		quitExitCode = EXIT_CODE_SHUTDOWN; 
		doQuit = true;
		quitServer(version + " (" + HOMEPAGE + ")");
		// System.exit() is called from onDisconnect() with the chosen exit code
	}

    public void restart() {
        if (LOG.isInfoEnabled()) {
            LOG.info(version + " shutdown - with exit code 9 (hoping for restart)");
        }
		quitExitCode = EXIT_CODE_RESTART; 
		doQuit = true;
        quitServer(version + " (" + HOMEPAGE + ") - Restarting ... ");
		// System.exit() is called from onDisconnect() with the chosen exit code
    }

    /**
     * Get the next configured nick.
     * @param currentNick used to find the next nick.
     * @return the nick after the current nick.
     */
    private String getNextNick(String currentNick) {
        String nextNick = "";

        int index = nicks.indexOf(currentNick);
        index++;

        if(index >= nicks.size()) {
          index = 0;
        }
        nextNick = (String) nicks.get(index);
        
        return nextNick;
    }

    /**
     * Get the next server to try connecting to.
     * @return the next configured server.
     * @see #getServer()
     */
    private IrcServer getNextServer() {
        if (this.servers.size() == 0) {
            return null;
        }
        if (++this.currentServer >= this.servers.size()) {
            this.currentServer = 0;
        }
        IrcServer server = (IrcServer) this.servers.get(this.currentServer);
        return server;
//        String lastTriedServer = getServer();
//        if(lastTriedServer != null) {
//            Iterator servs = cfgServers.iterator();
//            boolean foundLastTriedServer = false;
//            while(servs.hasNext()) {
//                Element serv = (Element) servs.next();
//                String retServer = serv.getAttributeValue("name");
//                if(foundLastTriedServer) {
//                    return retServer;
//                }
//               if(lastTriedServer.equalsIgnoreCase(retServer)) {
//                   foundLastTriedServer = true;
//               }
//          }
//          // if last tried server is last in list, it never gets to the "return retServer;" above - instead it gets here, and fallbacks (below) to returning the first server in the list
//        }
//        Element serv = (Element) cfgServers.get(0);
//        String retServer = serv.getAttributeValue("name");
//        
//        return retServer;
    }
    
    /**
     * Load the configured plugins.
     */
	private void loadPlugins() {
        List startupPlugins = new ArrayList(20);

        // load all of the plugins
        for (Iterator iter = plugins.values().iterator(); iter.hasNext();) {
            PPFPluginInfo pi = (PPFPluginInfo) iter.next();
            if (pi.isLoadAtStartup()) {
                String pluginName = pi.getName();
                startupPlugins.add(pluginName);
            }
        }

        for (Iterator tempIter = startupPlugins.iterator(); tempIter.hasNext();) {
            loadPlugin((String) tempIter.next());
        }

        for (Iterator tempIter = startupPlugins.iterator(); tempIter.hasNext();) {
            updatePluginLoadState((String) tempIter.next(), true);
        }

        if (LOG.isInfoEnabled()) {
            LOG.info("["+ getName() +"] "+ ppfResBun.getString(language, 
                "info_plugins_loaded", "All plugins loaded"));
        }
	}

    /**
     * Reload a plugin.
     * @param pluginName the name of the plugin to reload.
     * @return true if the plugin was successfully reloaded.
     */
    private boolean reloadPlugin(String pluginName) {
        unloadPlugin(pluginName);
        return loadPlugin(pluginName);
    }

    /**
     * Load a plugin.
     * @param pluginName the name of the plugin to load.
     * @return true if the plugin was successfully loaded.
     */
	private boolean loadPlugin(String pluginName) {
        boolean success = false;
        // get a handler to the plugins dir
        File file = new File(this.pluginsDir, pluginName);

        try {
            if (!file.exists() || !file.isDirectory()) {
                throw new IllegalStateException("Plugins directory not found: "
                        + file.getPath());
            }

            // get all the files in the dir
            String[] preSortedFiles = file.list();
            if (preSortedFiles == null || preSortedFiles.length == 0) {
                LOG.error(pluginName + ": plugin directory is empty");
            } else {
                // TODO: check what it does when a name is given but nothing exists
                // there
                // only load .JAR .ZIP and .class files
                String[] files = filterClassfiles(preSortedFiles);
                // build the full classpath
                URL[] urls = new URL[files.length + 1];
                urls[0] = new URL("file:" + this.pluginsDir
                        + File.separatorChar + pluginName + File.separatorChar);
                for (int f = 1; f < files.length + 1; f++) {
                    String classPath = this.pluginsDir + File.separatorChar
                            + pluginName + File.separatorChar + files[f - 1];
                    urls[f] = new URL("file:" + classPath);
                }

                // create a new classloader for each plugin with the needed classpath
                ClassLoader botCL = new URLClassLoader(urls, PPF.class
                        .getClassLoader());

                Thread.currentThread().setContextClassLoader(botCL);

                Class clazz = null;
                PPFPluginInfo pi = (PPFPluginInfo) plugins.get(pluginName);
                clazz = botCL.loadClass(pi.getClassName());

                // create a new instance of each plugin - this adds the plugin to the
                // plugin list
                Object obj = clazz.newInstance();

                PPFPlugin pluginInstance = (PPFPlugin) obj;
                pluginInstance.init(this);
                pi.setVersion(pluginInstance.getVersion());
                pi.setLoaded(true);
                plugins.remove(pluginName);
                plugins.put(pluginName, pi);

                loadedPlugins.put(pluginName, pluginInstance);
                
                success = true;

                if (LOG.isInfoEnabled()) {
                    LOG.info("["
                            + getName()
                            + "] "
                            + ppfResBun.getString(language,
                                    "info_plugin_loaded", "plugin loaded:") + " "
                            + pluginName);
                }
            }
        } catch (Exception e) {
            LOG.error("Error loading plugin: " + pluginName, e);
        }
        return success;
    }	

    /**
     * Unload a plugin
     * @param pluginName the name of the plugin to unload.
     */
    private void unloadPlugin(String pluginName) {
        if(loadedPlugins.containsKey(pluginName)) {
            PPFPlugin plug = (PPFPlugin) loadedPlugins.get(pluginName);
            try {
                plug.unload();
            } catch (Exception e) {
                LOG.error("Error while unloading plugin " + pluginName, e);
            }
            loadedPlugins.remove(pluginName);
            updatePluginLoadState(pluginName, false);
            if (LOG.isInfoEnabled()) {
                LOG.info("["+ getName() +"] "+ ppfResBun.getString(language, 
                        "info_plugin_unloaded", "INFO: plugin unloaded:") +" "+ pluginName);
            }
        }
    }

    /**
     * Update a plugin's internal load state.
     * @param pluginName the name of the plugin.
     * @param isLoaded whether's the plugin is now loaded or not.
     */
    private void updatePluginLoadState(String pluginName, boolean isLoaded) {
        if(plugins.containsKey(pluginName)) {
            PPFPluginInfo pi = (PPFPluginInfo) plugins.get(pluginName);
            pi.setLoaded(isLoaded);
            plugins.remove(pluginName);
            plugins.put(pluginName, pi);
        }
    }
    
    /**
     * Filter the specified file list for JAR, ZIP, CLASS and PROPERTIES files.
     * @param preSorted the file list to filter.
     * @return the filtered file list 
     */
	private String[] filterClassfiles(String[] preSorted) {
		int classCount = 0;
		// count
		for(int x=0; x < preSorted.length; x++) {
			if(preSorted[x].toUpperCase().endsWith(".JAR") ||
				preSorted[x].toUpperCase().endsWith(".ZIP") ||
				preSorted[x].toUpperCase().endsWith(".CLASS") ||
				preSorted[x].toUpperCase().endsWith(".PROPERTIES")) {
					
					classCount++;
			}
		}
		
		String[] sorted = new String[classCount];
		int insertCount = 0;
		// put into correctly sized array
		for(int x=0; x < preSorted.length; x++) {
			if(preSorted[x].toUpperCase().endsWith(".JAR") ||
				preSorted[x].toUpperCase().endsWith(".ZIP") ||
				preSorted[x].toUpperCase().endsWith(".CLASS") ||
                preSorted[x].toUpperCase().endsWith(".PROPERTIES")) {
					
					sorted[insertCount] = preSorted[x];
					insertCount++;
			}
		}
		return sorted;
	}

    protected void onNickChange(String oldNick, String login, String hostname, 
        String newNick) {
        
        if(isAuthed(oldNick, AUTH_ANY)) {
            // get the authed user object
            AuthedUser au = (AuthedUser) authedUsers.get(oldNick);
            // update the nick
            au.setUser(newNick);
            // remove the entry with the old nick
            authedUsers.remove(oldNick);
            // add the updated authed user object with the new nick
            authedUsers.put(newNick, au);             
        }
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onNickChange(oldNick, login, hostname, newNick);
            } catch (Exception e) {
                LOG.error("Error handling nick change in plugin " + plugin.getClass().getName(), e);
            }
        }
    }

    protected void onAction(String sender, String login, String hostname, 
        String target, String action) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onAction(sender, login, hostname, target, action);
            } catch (Exception e) {
                LOG.error("Error handling action in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
    
    protected void onChannelInfo(String channel, int userCount, String topic) {
    
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onChannelInfo(channel, userCount, topic);
            } catch (Exception e) {
                LOG.error("Error handling channel info in plugin " + plugin.getClass().getName(), e);
            }
        }
    }

    protected void onConnect() {

        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onConnect();
            } catch (Exception e) {
                LOG.error("Error handling connect in plugin " + plugin.getClass().getName(), e);
            }
        }
    } 

    protected void onDeop(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String recipient) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onDeop(channel, sourceNick, sourceLogin, sourceHostname,
                        recipient);
            } catch (Exception e) {
                LOG.error("Error handling deop in plugin " + plugin.getClass().getName(), e);
            }
        }
    } 

    protected void onDeVoice(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String recipient) {
                    
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onDeVoice(channel, sourceNick, sourceLogin,
                        sourceHostname, recipient);
            } catch (Exception e) {
                LOG.error("Error handling devoice in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onFileTransferFinished(DccFileTransfer transfer, Exception e) {
    
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            plugin.onFileTransferFinished(transfer, e);
        }
    } 

    protected void onFinger(String sourceNick, String sourceLogin, 
        String sourceHostname, String target) {

        super.onFinger(sourceNick, sourceLogin, sourceHostname, target);
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin
                        .onFinger(sourceNick, sourceLogin, sourceHostname,
                                target);
            } catch (Exception e) {
                LOG.error("Error handling finger in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
    
    protected void onIncomingChatRequest(DccChat chat) {
    
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onIncomingChatRequest(chat);
            } catch (Exception e) {
                LOG.error("Error handling DCC chat request in plugin " + plugin.getClass().getName(), e);
            }
        }
    }                 

    protected void onIncomingFileTransfer(DccFileTransfer transfer) {
    
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onIncomingFileTransfer(transfer);
            } catch (Exception e) {
                LOG.error("Error handling incoming file transfer in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
    
    protected void onInvite(String targetNick, String sourceNick, String sourceLogin, 
        String sourceHostname, String channel) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onInvite(targetNick, sourceNick, sourceLogin,
                        sourceHostname, channel);
            } catch (Exception e) {
                LOG.error("Error handling invite in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
                
    protected void onMode(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String mode) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onMode(channel, sourceNick, sourceLogin, sourceHostname,
                        mode);
            } catch (Exception e) {
                LOG.error("Error handling mode in plugin " + plugin.getClass().getName(), e);
            }
        }
    } 
        
    protected void onNotice(String sourceNick, String sourceLogin, String sourceHostname, 
        String target, String notice) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onNotice(sourceNick, sourceLogin, sourceHostname,
                        target, notice);
            } catch (Exception e) {
                LOG.error("Error handling notice in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
    
    protected void onOp(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String recipient) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onOp(channel, sourceNick, sourceLogin, sourceHostname,
                        recipient);
            } catch (Exception e) {
                LOG.error("Error handling op in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onPing(String sourceNick, String sourceLogin, 
        String sourceHostname, String target, String pingValue) {
        
        super.onPing(sourceNick, sourceLogin, sourceHostname, target, pingValue);
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onPing(sourceNick, sourceLogin, sourceHostname, target,
                        pingValue);
            } catch (Exception e) {
                LOG.error("Error handling ping in plugin " + plugin.getClass().getName(), e);
            }
        }
    }

    protected void onRemoveChannelBan(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname, String hostmask) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onRemoveChannelBan(channel, sourceNick, sourceLogin,
                        sourceHostname, hostmask);
            } catch (Exception e) {
                LOG.error("Error handling remove channel ban in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onRemoveChannelKey(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname, String key) {
        
        if(currentChannels.containsKey(channel)) {
            currentChannels.remove(channel);
        }
        currentChannels.put(channel, "");

        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onRemoveChannelKey(channel, sourceNick, sourceLogin,
                        sourceHostname, key);
            } catch (Exception e) {
                LOG.error("Error handling remove channel key in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onRemoveChannelLimit(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onRemoveChannelLimit(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling remove channel limit in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onRemoveInviteOnly(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onRemoveInviteOnly(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling remove invite only in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onRemoveModerated(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onRemoveModerated(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling remove moderated in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onRemoveNoExternalMessages(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onRemoveNoExternalMessages(channel, sourceNick,
                        sourceLogin, sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling remove external messages in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onRemovePrivate(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onRemovePrivate(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling remove private in plugin " + plugin.getClass().getName(), e);
            }
        }
    } 
 
    protected void onRemoveSecret(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onRemoveSecret(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling remove secret in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onRemoveTopicProtection(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onRemoveTopicProtection(channel, sourceNick,
                        sourceLogin, sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling remove topic protection in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onServerPing(String response) {
        
        super.onServerPing(response);
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onServerPing(response);
            } catch (Exception e) {
                LOG.error("Error handling server ping in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
    
    protected void onServerResponse(int code, String response) {
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onServerResponse(code, response);
            } catch (Exception e) {
                LOG.error("Error handling server response in plugin " + plugin.getClass().getName(), e);
            }
        }
    }

    protected void onSetChannelKey(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname, String key) {
        
        if(currentChannels.containsKey(channel)) {
            currentChannels.remove(channel);
        }
        currentChannels.put(channel, key);
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onSetChannelKey(channel, sourceNick, sourceLogin,
                        sourceHostname, key);
            } catch (Exception e) {
                LOG.error("Error handling set channel key in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onSetChannelLimit(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname, int limit) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onSetChannelLimit(channel, sourceNick, sourceLogin,
                        sourceHostname, limit);
            } catch (Exception e) {
                LOG.error("Error handling set channel key in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onSetInviteOnly(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onSetInviteOnly(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling set invite only in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onSetModerated(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onSetModerated(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling set moderated in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onSetNoExternalMessages(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onSetNoExternalMessages(channel, sourceNick,
                        sourceLogin, sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling set no external messages in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onSetPrivate(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onSetPrivate(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling set private in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onSetSecret(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onSetSecret(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling set secret in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onSetTopicProtection(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onSetTopicProtection(channel, sourceNick, sourceLogin,
                        sourceHostname);
            } catch (Exception e) {
                LOG.error("Error handling set topic protection in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onTime(String sourceNick, String sourceLogin, 
        String sourceHostname, String target) {
            
        super.onTime(sourceNick, sourceLogin, sourceHostname, target);
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onTime(sourceNick, sourceLogin, sourceHostname, target);
            } catch (Exception e) {
                LOG.error("Error handling time in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
    
    protected void onTopic(String channel, String topic, String setBy, long date, boolean changed) {
        //topic things
        ChannelInfo ci;
        if (liveInfo.containsKey(channel.toLowerCase())) {
            ci = (ChannelInfo) liveInfo.remove(channel.toLowerCase());
        } else {
            ci = new ChannelInfo();
        }
        ci.setTopic(topic);
        liveInfo.put(channel.toLowerCase(), ci);

        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onTopic(channel, topic, setBy, date, changed);
            } catch (Exception e) {
                LOG.error("Error handling topic in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onUnknown(String line) {
    
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onUnknown(line);
            } catch (Exception e) {
                LOG.error("Error handling unknown in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
    
    protected void onUserList(String channel, User[] users) {
    
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onUserList(channel, users);
            } catch (Exception e) {
                LOG.error("Error handling user list in plugin " + plugin.getClass().getName(), e);
            }
        }
    } 
  
    protected void onUserMode(String targetNick, String sourceNick, 
        String sourceLogin, String sourceHostname, String mode) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onUserMode(targetNick, sourceNick, sourceLogin,
                        sourceHostname, mode);
            } catch (Exception e) {
                LOG.error("Error handling user mode in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onVersion(String sourceNick, String sourceLogin, 
        String sourceHostname, String target) {
        
        super.onVersion(sourceNick, sourceLogin, sourceHostname, target);
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onVersion(sourceNick, sourceLogin, sourceHostname,
                        target);
            } catch (Exception e) {
                LOG.error("Error handling version in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
        
    protected void onVoice(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String recipient) {
        
        Iterator iter = loadedPlugins.values().iterator();
        while(iter.hasNext()) {
            PPFPlugin plugin = (PPFPlugin) iter.next();
            try {
                plugin.onVoice(channel, sourceNick, sourceLogin,
                        sourceHostname, recipient);
            } catch (Exception e) {
                LOG.error("Error handling voice in plugin " + plugin.getClass().getName(), e);
            }
        }
    }
    
    /**
     * Parse command line parameters and start the bot.
     * @param args command line parameters.
     */
   	public static void main(String args[]) {
        String cfgFilename = DEFAULT_CONFIG_FILENAME;
        // the for loop is just for future extensions
        for (int i = 0; i < args.length; i++) {
            String arg = args[i];
            if (arg.equals("-h") || arg.equals("--help")) {
                System.out.println(getUsage());
                return;
            } else if (arg.equals("-V") || arg.equals("--version")) {
                System.out.println(fullVersion);
                return;
            } else {
                cfgFilename = arg;
                break;
            }
        }
        try {
            PPF bot = new PPF();
            bot.configure(cfgFilename, false);
            bot.run();
        } catch (Throwable t) {
            if (LOG.isDebugEnabled()) {
                LOG.fatal("Fatal error: " + t.getLocalizedMessage(), t);
            }
            System.exit(EXIT_CODE_FATAL);
        }
	}

    /**
     * Usage help message for PPF explaining the command line parameters.
     * @return PPF usage
     */
    public static String getUsage() {
        return "Usage: java " + PPF.class.getName() + "[options] [config file]"
                    + "\n    where options are:"
                    + "\n    -h    --help    Print this help and exit"
                    + "\n    -V    --verbose Print version and exit"
                    + "\n\nIf no config file is specified, the default of PPFConfig.xml is used";
    }

    /**
     * If a user has set a language return that language, else return the bot's
     * configured default language.
     * @param nick the user's nick.
     * @return the language ISO code (en, de, etc.)
     */
    public String getUserLang(String nick) {
        if(userLangs.containsKey(nick)) {
            return (String) userLangs.get(nick);
        }
        // default to the language that the bot is configured for
        return language;
    }
    
    /**
     * The channels the bot is currently on.
     * @return the bot's channels.
     */
    public Map getCurrentChannels() {
        return currentChannels;
    }

    /**
     * Simulate a private message. This makes PPF and the plugins handle the
     * message, but does not send it to the server.
     * @see #onPrivateMessage(java.lang.String,java.lang.String,java.lang.String,java.lang.String)
     */
    public void digestPrivateMessage(String sender, String login, String hostname,
        String message) {
            
        onPrivateMessage(sender, login, hostname, message);
    }

    /**
     * Simulate a channel message. This makes PPF and the plugins handle the
     * message, but does not send it to the server.
     * @see #onMessage(java.lang.String,java.lang.String,java.lang.String,java.lang.String)
     */
    public void digestMessage(String channel, String sender, String login,
        String hostname, String message) {

        onMessage(channel, sender, login, hostname, message);
    }

    private String checkPluginUpdates(String pluginName, String currentVersion) {
        String urlinfo = "http://www.deaded.com/software/updates/plugins/"+ pluginName + "/version.txt";
        
        HttpURLConnection huc = null;
        InputStreamReader is = null;
        BufferedReader reader = null;
        String line = "";
        try {
            URL url = new URL(urlinfo);
            huc = (HttpURLConnection) url.openConnection();            
            huc.setRequestMethod("GET");
            //TODO include PPF version in user agent? ...just for the stats
            huc.setRequestProperty("User-Agent", "Mozilla/4.0 (compatible; PircBot; JVM)");                        
            huc.setRequestProperty("Pragma", "no-cache");
            huc.connect();
            is = new InputStreamReader(huc.getInputStream());
            reader = new BufferedReader(is);
        
            // only a one liner
            line = reader.readLine();
                
        } catch (Exception e) {
            LOG.error("Error querying for updates for plugin " + pluginName, e);
        } finally {
            if(reader != null) {
                try {
                    reader.close();
                } catch (IOException e1) {
                    // messed up at this point anyway 
                    LOG.error("Error closing reader for " + urlinfo, e1);
                }
            }
            if(is != null) {
                try {
                    is.close();
                } catch (IOException e1) {
                    // messed up at this point anyway 
                    LOG.error("Error closing input stream for " + urlinfo, e1);
                }
            }
            if(huc != null) {
                huc.disconnect();
            }
        }
        
        return line;
    }

    private void downloadPluginUpdate(String pluginName, String downloadTo) {
        String urlinfo = "http://www.deaded.com/software/updates/plugins/"+ pluginName + "/"+ pluginName +".jar";
        String outputfile = downloadTo +"/"+ pluginName +".jar";
        
        FileOutputStream fos = null;
        InputStream is = null;
        try {        
            URL url = new URL(urlinfo);
        
            // Copy resource to local file, use remote file
            // if no local file name specified
            // Print info about resource
        
            is = url.openStream();
            fos = new FileOutputStream(outputfile);
        
            int oneChar, count=0;
            while ((oneChar=is.read()) != -1) {
              fos.write(oneChar);
              count++;
            }

        } catch (MalformedURLException e) {
            LOG.error("Not a valid address: " + urlinfo);
        } catch (IOException e) {
            LOG.error("Error downloading updates for plugin " + pluginName, e);
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException e) {
                    LOG.error("Error closing input stream for " + urlinfo, e);
                }
            }
            if(fos != null) {
                try {
                    fos.close();
                } catch (IOException e1) {
                    // messed up at this point anyway 
                    LOG.error("Error file output stream for " + outputfile, e1);
                }
            }
        } 
    }
    
    public Map getLoadedPlugins() {
        return loadedPlugins;
    }
    
    public Map getCfgPlugins() {
        return plugins;
    }
    
    public String urlEncode(String string) {
        try {
            return URLEncoder.encode(string, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        
        return string;
    }

    /**
     * @return Returns the cfgFilename.
     */
    public String getCfgFilename() {
        return cfgFilename;
    }
    

    /**
     * @param cfgFilename The cfgFilename to set.
     */
    public void setCfgFilename(String cfgFilename) {
        this.cfgFilename = cfgFilename;
    }
    

    /**
     * @return Returns the cfgAdminPassword.
     */
    public String getAdminPassword() {
        return adminPassword;
    }
    

    /**
     * @param cfgAdminPassword The cfgAdminPassword to set.
     */
    public void setAdminPassword(String cfgAdminPassword) {
        this.adminPassword = cfgAdminPassword;
    }
    


    /**
     * @return Returns the cfgCommandHelp.
     */
    public String getCommandHelp() {
        return commandHelp;
    }
    

    /**
     * @param cfgCommandHelp The cfgCommandHelp to set.
     */
    public void setCommandHelp(String cfgCommandHelp) {
        this.commandHelp = cfgCommandHelp;
    }
    

    /**
     * @return Returns the cfgCommandSetLang.
     */
    public String getCommandSetLang() {
        return commandSetLang;
    }
    

    /**
     * @param cfgCommandSetLang The cfgCommandSetLang to set.
     */
    public void setCommandSetLang(String cfgCommandSetLang) {
        this.commandSetLang = cfgCommandSetLang;
    }
    

    /**
     * @return Returns the cfgEncoding.
     */
    public String getEncoding() {
        return encoding;
    }
    

    /**
     * @param cfgEncoding The cfgEncoding to set.
     */
    public void setEncoding(String cfgEncoding) {
        this.encoding = cfgEncoding;
    }
    

    /**
     * @return Returns the cfgIdentNick.
     */
    public String getIdentNick() {
        return identNick;
    }
    

    /**
     * @param cfgIdentNick The cfgIdentNick to set.
     */
    public void setIdentNick(String cfgIdentNick) {
        this.identNick = cfgIdentNick;
    }
    

    /**
     * @return Returns the cfgLanguage.
     */
    public String getLanguage() {
        return language;
    }
    

    /**
     * @param cfgLanguage The cfgLanguage to set.
     */
    public void setLanguage(String cfgLanguage) {
        this.language = cfgLanguage;
    }
    

    /**
     * @return Returns the cfgMasterPassword.
     */
    public String getMasterPassword() {
        return masterPassword;
    }
    

    /**
     * @param cfgMasterPassword The cfgMasterPassword to set.
     */
    public void setMasterPassword(String cfgMasterPassword) {
        this.masterPassword = cfgMasterPassword;
    }
    

    /**
     * @return Returns the cfgOnJoinMessage.
     */
    public String getOnJoinMessage() {
        return onJoinMessage;
    }
    

    /**
     * @param cfgOnJoinMessage The cfgOnJoinMessage to set.
     */
    public void setOnJoinMessage(String cfgOnJoinMessage) {
        this.onJoinMessage = cfgOnJoinMessage;
    }
    

    /**
     * @return Returns the cfgRejoinOnKick.
     */
    public boolean isRejoinOnKick() {
        return rejoinOnKick;
    }
    

    /**
     * @param cfgRejoinOnKick The cfgRejoinOnKick to set.
     */
    public void setRejoinOnKick(boolean cfgRejoinOnKick) {
        this.rejoinOnKick = cfgRejoinOnKick;
    }
    

    /**
     * @return Returns the cfgTrustedPassword.
     */
    public String getTrustedPassword() {
        return trustedPassword;
    }
    

    /**
     * @param cfgTrustedPassword The cfgTrustedPassword to set.
     */
    public void setTrustedPassword(String cfgTrustedPassword) {
        this.trustedPassword = cfgTrustedPassword;
    }
    

    /**
     * Add a server.<br>
     * Should be called before connecting.
     * @param server the server to add.
     */
    public void addServer(IrcServer server) {
        this.servers.add(server);
    }
    
    /**
     * Add a nick for the bot.<br>
     * Should be called before connecting.
     * @param nick the nick to add.
     */
    public void addNick(String nick) {
        this.nicks.add(nick);
    }
    
    /**
     * Add a channel to join.<br>
     * Should be called before connecting,
     * @param channel the channel to add.
     */
    public void addChannel(IrcChannel channel) {
        this.channels.add(channel);
    }

    /**
     * Register a plugin, so it can be loaded later.
     * @param pluginInfo
     */
    public void addPluginInfo(PPFPluginInfo pluginInfo) {
        this.plugins.put(pluginInfo.getName(), pluginInfo);
        boolean loaded = loadedPlugins.containsKey(pluginInfo.getName());
        pluginInfo.setLoaded(loaded);
    }
    
    public void setDccPorts(String strDCCPorts) {
        if(strDCCPorts.length() > 0) {
            // config eg: 2000, 2500-2600, 2900
            String[] parts = strDCCPorts.split(",");
            List allowedPorts = new ArrayList();
            for(int p = 0; p < parts.length; p++) {
                String piece = parts[0];
                if(piece.indexOf("-") > 0) {
                    String one = piece.substring(0, piece.indexOf("-"));
                    String two = piece.substring(one.length()+1);
                    allowedPorts.add(one);
                    allowedPorts.add(two);
                } else {
                    allowedPorts.add(piece);
                }
            }
            try {
                cfgDCCPorts = new int[allowedPorts.size()];
                for(int a = 0; a < allowedPorts.size(); a++) {
                    cfgDCCPorts[a] = Integer.valueOf((String) allowedPorts.get(a)).intValue();
                    setDccPorts(cfgDCCPorts);
                }
            } catch (NumberFormatException e) {
                LOG.warn("Invalid or no value for dccPorts, defaulting to no specific ports being set");
                setDccPorts((int[])null);                    
            }
        }
    }

    /**
     * @return Returns the maxConnectionAttempts.
     */
    public int getMaxConnectionAttempts() {
        return maxConnectionAttempts;
    }
    

    /**
     * @param maxConnectionAttempts The maxConnectionAttempts to set.
     */
    public void setMaxConnectionAttempts(int maxConnectionAttempts) {
        this.maxConnectionAttempts = maxConnectionAttempts;
    }
    

    /**
     * @return Returns the retryDelay in seconds.
     */
    public long getRetryDelay() {
        return retryDelay;
    }
    

    /**
     * Set in milliseconds and multiplied by 1000 for later use.
     * 
     * @param retryDelay The retryDelay to set.
     */
    public void setRetryDelay(long retryDelay) {
        this.retryDelay = retryDelay * 1000;
    }
    

    /**
     * @return Returns the startWebServer.
     */
    public boolean isStartWebServer() {
        return startWebServer;
    }
    

    /**
     * @param startWebServer The startWebServer to set.
     */
    public void setStartWebServer(boolean startWebServer) {
        this.startWebServer = startWebServer;
    }
    

    /**
     * @return Returns the webServerDocumentRoot.
     */
    public String getWebServerDocumentRoot() {
        return webServerDocumentRoot;
    }
    

    /**
     * @param webServerDocumentRoot The webServerDocumentRoot to set.
     */
    public void setWebServerDocumentRoot(String webServerDocumentRoot) {
        this.webServerDocumentRoot = webServerDocumentRoot;
    }
    

    /**
     * @return Returns the webServerPort.
     */
    public int getWebServerPort() {
        return webServerPort;
    }
    

    /**
     * @param webServerPort The webServerPort to set.
     */
    public void setWebServerPort(int webServerPort) {
        this.webServerPort = webServerPort;
    }

    /******************
     * topic settings *
     ******************/
    private ChannelInfo getChannelInfo(String channel) {
        return (ChannelInfo)liveInfo.get(channel.toLowerCase());
    }

    /**
     * Returns the topic with irc codes for formatting and colors.
     *
     * @return topic
     */
    public String getTopic(String channel) {
        return getChannelInfo(channel).getCodeTopic();
    }

	public boolean isHidePingPong() {
		return hidePingPong;
	}

	public void setHidePingPong(boolean hidePingPong) {
		this.hidePingPong = hidePingPong;
	}

	public void setPluginsDir(String pluginsDir) {
		this.pluginsDir=pluginsDir;
	}

}
