/* 
Copyright Paul James Mutton, 2001-2004, http://www.jibble.org/

This file is part of Mini Wegb Server / SimpleWebServer.

This software is dual-licensed, allowing you to choose between the GNU
General Public License (GPL) and the www.jibble.org Commercial License.
Since the GPL may be too restrictive for use in a proprietary application,
a commercial license is also provided. Full license information can be
found at http://www.jibble.org/licenses/

$Author: deaded $
$Id: RequestThread.java,v 1.2 2005/03/28 08:47:36 deaded Exp $

*/

package net.sourceforge.ppf.simplewebserver;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.net.Socket;
import java.net.URLDecoder;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;

import net.sourceforge.ppf.PPF;

import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;
import org.apache.xerces.impl.dv.util.Base64;

/**
 * Copyright Paul Mutton
 * http://www.jibble.org/
 *
 */
public class RequestThread extends Thread {
    
    public RequestThread(Socket socket, File rootDir, PPF bot) {
    	_bot = bot;
        _socket = socket;
        _rootDir = rootDir;
    }
/*    
    public RequestThread(Socket socket, File rootDir) {
        _socket = socket;
        _rootDir = rootDir;
    }
*/
    private static void sendHeader(BufferedOutputStream out, int code, String contentType, long contentLength, long lastModified) throws IOException {
    	sendHeader(out, code, contentType, contentLength, lastModified, null);
    }

   	private static void sendHeader(BufferedOutputStream out, int code, String contentType, long contentLength, long lastModified, String header) throws IOException {
        out.write(("HTTP/1.0 " + code + " OK\r\n" + 
                   "Date: " + new Date().toString() + "\r\n" +
                   "Server: PPFWebServer/1.0\r\n" +
                   "Content-Type: " + contentType + "\r\n" +
                   "Expires: Thu, 01 Dec 1994 16:00:00 GMT\r\n" +
                   ((contentLength != -1) ? "Content-Length: " + contentLength + "\r\n" : "") +
                   "Last-modified: " + new Date(lastModified).toString() + "\r\n" +
				   (header != null ? header + "\r\n" : "") +
                   "\r\n").getBytes());
    }
    
    private static void sendError(BufferedOutputStream out, int code, String message) throws IOException {
        message = message + "<hr>" + SimpleWebServer.VERSION;
        sendHeader(out, code, "text/html", message.length(), System.currentTimeMillis());
        out.write(message.getBytes());
        out.flush();
        out.close();
    }
    
    public void run() {
        InputStream reader = null;
        try {
            _socket.setSoTimeout(30000);
            BufferedReader in = new BufferedReader(new InputStreamReader(_socket.getInputStream()));
            BufferedOutputStream out = new BufferedOutputStream(_socket.getOutputStream());
            
            String request = in.readLine();
            if (request == null || !request.startsWith("GET ") || !(request.endsWith(" HTTP/1.0") || request.endsWith("HTTP/1.1"))) {
                // Invalid request type (no "GET")
                sendError(out, 500, "Invalid Method.");
                return;
            }

			//Read in and store all the headers.
			HashMap headers = new HashMap();
			String line = null;
			while ((line = in.readLine()) != null) {
			    line = line.trim();
			    if (line.equals("")) {
			        break;
			    }
			    int colonPos = line.indexOf(":");
			    if (colonPos > 0) {
			        String key = line.substring(0, colonPos);
			        String value = line.substring(colonPos + 1);
			        headers.put(key, value.trim());
			    }
			}

            String path = request.substring(4, request.indexOf("HTTP")).trim();
			
			String user = "";
			// TODO:check if path needs authenticating
			if(path.startsWith("/") && !path.equals("/favicon.ico")) {
				// yes it does
				boolean authed = false;
				String authorization = (String)headers.get("Authorization");
				if(authorization != null) {
					if(authorization.startsWith("Basic ")) {
						String base64 = authorization.substring(6);
						String challenge = new String(Base64.decode(base64));
						int colonIdx = challenge.indexOf(':');
						user = challenge.substring(0, colonIdx);
						String pass = challenge.substring(colonIdx+1);
						authed = _bot.verifyAuthentication(user, pass, _socket.getInetAddress().getHostName());
					}
				}
				
				if(!authed) {
					VelocityEngine ve = new VelocityEngine();
					ve.init();
					ve.clearProperty("resource.manager.cache.class"); 
					VelocityContext context = new VelocityContext();
					Template t = ve.getTemplate("web-templates/notauthed.vm");
					StringWriter writer = new StringWriter();
					
					context.put("title", "Error - Access Denied");
					
					t.merge(context, writer);
		
					sendHeader(out, 401, "text/html", writer.toString().length(), 0,
				    		   "WWW-Authenticate: Basic realm=\"PPFWorld\"");
				    out.write(writer.toString().getBytes()); 
				    out.close();
				    return;
				}
			}
			
			final String WEB_ROOT = "/botControl";  // TODO: this will come as config from bot config
			// anything in the admin control panel
			if(path.startsWith(WEB_ROOT)) {
                // setup velocity
				VelocityEngine ve = new VelocityEngine();
				ve.init();
				ve.clearProperty("resource.manager.cache.class"); 
				VelocityContext context = new VelocityContext();
				String templateToUse = "web-templates/index.vm"; // index by default
				StringWriter writer = new StringWriter();

				context.put("webroot", WEB_ROOT);
				context.put("title", _bot.getName() +" - Admin Control Panel");
				context.put("user", user);
			    SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy HH:mm");
			    Calendar cal = Calendar.getInstance(); // today
				context.put("datetime", sdf.format(cal.getTime()));
				context.put("bot", _bot);

                int queryLoc = path.indexOf("?");
                boolean hasQuery = false;
                String query = "";
                if(queryLoc >= 0) {
                    hasQuery = true;
                    query = path.substring(queryLoc + 1);
                }

                HashMap parameters = new HashMap();
                String[] pairs = query.split("&");
                for(int i=0; i<pairs.length; ++i) {
                    int idx = pairs[i].indexOf('=');
                    if(idx == -1)
                        continue;
                    String key = URLDecoder.decode(pairs[i].substring(0, idx), "UTF-8");
                    String value = URLDecoder.decode(pairs[i].substring(idx+1), "UTF-8");
                    parameters.put(key, value);
                }
                
				if(!(path.equals(WEB_ROOT))) { // anything but index
	
                    if(path.startsWith(WEB_ROOT +"/channel")) {
                        templateToUse = "web-templates/channel.vm";
                        context.put("channel", (String) parameters.get("channel"));
                    } else if(path.startsWith(WEB_ROOT +"/plugins")) {
                        templateToUse = "web-templates/plugins.vm";
                    }

				}
                _bot.fromWeb(user, _socket.getInetAddress().getHostName(), parameters);

				// now process the template
				Template t = ve.getTemplate(templateToUse);
				t.merge(context, writer);
				
				sendHeader(out, 200, "text/html; charset=utf-8", writer.toString().length(), 0);
				out.write(writer.toString().getBytes());

			// if not using the admin control panel, go for public web server usage
			} else {
			
	            File file = new File(_rootDir, URLDecoder.decode(path, "UTF-8")).getCanonicalFile();
	            if (file.isDirectory()) {
	                // Check to see if there is an index file in the directory.
	                File indexFile = new File(file, "index.html");
	                if (indexFile.exists() && !indexFile.isDirectory()) {
	                    file = indexFile;
	                }
	            }
	
	            if (!file.toString().startsWith(_rootDir.toString())) {
	                // Uh-oh, it looks like some lamer is trying to take a peek
	                // outside of our web root directory.
	                sendError(out, 403, "Permission Denied.");
	            }
	            else if (!file.exists()) {
	                // The file was not found.
	                sendError(out, 404, "File Not Found.");
	            }
	            else if (file.isDirectory()) {
	                // print directory listing
	                if (!path.endsWith("/")) {
	                    path = path + "/";
	                }
	                File[] files = file.listFiles();
	                sendHeader(out, 200, "text/html", -1, System.currentTimeMillis());
	                String title = "Index of " + path;
	                out.write(("<html><head><title>" + title + "</title></head><body><h3>Index of " + path + "</h3><p>\n").getBytes());
	                for (int i = 0; i < files.length; i++) {
	                    file = files[i];
	                    String filename = file.getName();
	                    String description = "";
	                    if (file.isDirectory()) {
	                        description = "&lt;DIR&gt;";
	                    }
	                    out.write(("<a href=\"" + path + filename + "\">" + filename + "</a> " + description + "<br>\n").getBytes());
	                }
	                out.write(("</p><hr><p>" + SimpleWebServer.VERSION + "</p></body><html>").getBytes());
	            }
	            else {
	                reader = new BufferedInputStream(new FileInputStream(file));
	            
	                String contentType = (String)SimpleWebServer.MIME_TYPES.get(SimpleWebServer.getExtension(file));
	                if (contentType == null) {
	                    contentType = "application/octet-stream";
	                }
	                
	                sendHeader(out, 200, contentType, file.length(), file.lastModified());
	                
	                byte[] buffer = new byte[4096];
	                int bytesRead;
	                while ((bytesRead = reader.read(buffer)) != -1) {
	                    out.write(buffer, 0, bytesRead);
	                }
	                reader.close();
	            }
			}
            out.flush();
            out.close();
        }
        catch (IOException e) {
            if (reader != null) {
                try {
                    reader.close();
                }
                catch (Exception anye) {
                    // Do nothing.
                }
            }
        } catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
    }
    
    private File _rootDir;
    private Socket _socket;
    private PPF _bot;
    
}