package net.sourceforge.ppf.util;

import org.jibble.pircbot.Colors;

import java.util.ArrayList;


/**
 * Misc. String utilities
 *
 * @author DeadEd
 * @author 2Fast2BCn
 */
public abstract class StringUtil {
    public static String BLACK = "$BLACK";
    public static String BLUE = "$BLUE";
    public static String BROWN = "$BROWN";
    public static String CYAN = "$CYAN";
    public static String DARK_BLUE = "$DARK_BLUE";
    public static String DARK_GRAY = "$DARK_GRAY";
    public static String DARK_GREEN = "$DARK_GREEN";
    public static String GREEN = "$GREEN";
    public static String LIGHT_GRAY = "$LIGHT_GRAY";
    public static String MAGENTA = "$MAGENTA";
    public static String OLIVE = "$OLIVE";
    public static String PURPLE = "$PURPLE";
    public static String RED = "$RED";
    public static String TEAL = "$TEAL";
    public static String WHITE = "$WHITE";
    public static String YELLOW = "$YELLOW";

    public static String BOLD = "$BOLD";
    public static String NORMAL = "$NORMAL";
    public static String REVERSE = "$REVERSE";
    public static String UNDERLINE = "$UNDERLINE";

    public static String BACKGROUND = "$BG";

    public static String COLOR_SIGNS[] = {
        WHITE, BLACK, DARK_BLUE,
        DARK_GREEN, RED, BROWN,
        PURPLE, OLIVE, YELLOW,
        GREEN, TEAL, CYAN,
        BLUE, MAGENTA, DARK_GRAY,
        LIGHT_GRAY
    };
    public static String COLOR_CODES[] = {
        Colors.WHITE, Colors.BLACK, Colors.DARK_BLUE,
        Colors.DARK_GREEN, Colors.RED, Colors.BROWN,
        Colors.PURPLE, Colors.OLIVE, Colors.YELLOW,
        Colors.GREEN, Colors.TEAL, Colors.CYAN,
        Colors.BLUE, Colors.MAGENTA, Colors.DARK_GRAY,
        Colors.LIGHT_GRAY
    };
    public static String COLOR_HTML[] = {
        "#FFFFFF", "#000000", "#00007F",
        "#009300", "#FF0000", "#7F0000",
        "#9C009C", "#FC7F00", "#FFFF00",
        "#00FC00", "#00FFFF", "#009393",
        "#0000FC", "#FF00FF", "#7F7F7F",
        "#D2D2D2"
    };

    public static String FORMATTING_SIGNS[] = {
        BOLD, NORMAL, REVERSE, UNDERLINE
    };
    public static String FORMATTING_CODES[] = {
        Colors.BOLD, Colors.NORMAL, Colors.REVERSE, Colors.UNDERLINE
    };
    public static String FORMATTING_HTML[] = {
        "", "</font>", "", ""
    };

    private static ArrayList allColorCombinations;
    private static ArrayList allFormattingCombinations;


    static {
        allColorCombinations = new ArrayList();
        for (int outerLoop = 0; outerLoop < COLOR_SIGNS.length; outerLoop++) {
            for (int innerLoop = 0; innerLoop < COLOR_SIGNS.length; innerLoop++) {
                StringFormatRelation stringFormatRel = new StringFormatRelation(combineForeAndBackgroundToHTMLCode(outerLoop, innerLoop),
                        COLOR_SIGNS[outerLoop] + BACKGROUND + COLOR_SIGNS[innerLoop],
                        combineForeAndBackgroundToColorCode(outerLoop, innerLoop));
                allColorCombinations.add(stringFormatRel);
            }
        }
        for (int i = 0; i < COLOR_SIGNS.length; i++) {
            StringFormatRelation stringFormatRel = new StringFormatRelation("<font color=\"" + COLOR_HTML[i] + "\">",
                    COLOR_SIGNS[i],
                    COLOR_CODES[i]);
            allColorCombinations.add(stringFormatRel);
        }

        allFormattingCombinations = new ArrayList();
        for (int i = 0; i < FORMATTING_SIGNS.length; i++) {
            StringFormatRelation stringFormatRel = new StringFormatRelation(FORMATTING_HTML[i],
                    FORMATTING_SIGNS[i],
                    FORMATTING_CODES[i]);
            allFormattingCombinations.add(stringFormatRel);
        }
    }

    /*********************
     * DeadEd's methodes *
     *********************/

    /**
     * If the channelName does not start with a known prefix ('#' or '&amp;')
     * prepend it with a hash '#'.
     *
     * @param channelName the channel name
     * @return The channelName, prepended with a hash if it doesn't start with a
     *         known prefix.
     */
    public static String prependHash(String channelName) {
        if (channelName.startsWith("#") || channelName.startsWith("&")) {
            return channelName;
        } else {
            return "#" + channelName;
        }
    }

    /**
     * builds the usual nick!login@hostname string.
     */
    public static String formatUser(String nick, String login, String hostname) {
        return nick + "!" + login + "@" + hostname;
    }



    /************************
     * 2Fast2BCn's methodes *
     ************************/
    /**
     * removes some HTML tags and replaces some codes
     * <p/>
     * todo complete... or find better dynamic way...
     * todo I'm not happy with this code...
     * todo we might change <b> into IRC code for bold
     *
     * @param str
     * @return
     */
    public final static String replaceHTMLTags(String str) {
        return str.trim().replaceAll("&amp;", "&").replaceAll("&#39;", "'").replaceAll("&#146;", "'").replaceAll("&#233;", "").replaceAll("&#34;", "\"").replaceAll("&euml;", "").replaceAll("&acute;", "'").replaceAll("&rsquo;", "'").replaceAll("&#146;", "'").replaceAll("&#233;", "").replaceAll("</?[0-9a-zA-Z ]?[0-9a-zA-Z ]?[0-9a-zA-Z ][0-9a-zA-Z ]?>", "").replaceAll("&#?[a-zA-Z0-9 ]?[a-zA-Z0-9 ]?[a-zA-Z0-9 ]?[a-zA-Z0-9 ]?[a-zA-Z0-9 ]?[a-zA-Z0-9 ];", " ");
    }

    /**
     * Replaces all occurences of some special variables with their color/formatting
     * code in IRC format.
     * <p/>
     * The variables are in the form $VARIABLE, where VARIABLE is the name of one
     * of the constants in {@link org.jibble.pircbot.Colors}.
     *
     * @param original
     * @return
     */
    public static String replaceUTF8ToCodes(String original) {
        return replaceFormattingUTF8ToCode(replaceColorsUTF8ToCodes(original));
    }

    /**
     * Replaces all occurences of some special variables with their color/formatting
     * code in HTML format.
     * <p/>
     * todo implement (close all opened color codes... / backgroundcolors)
     * <p/>
     * The variables are in the form $VARIABLE, where VARIABLE is the name of one
     * of the constants in {@link org.jibble.pircbot.Colors}.
     *
     * @param original
     * @return
     */
    public static String replaceUTF8ToHTML(String original) {
        return replaceFormattingUTF8ToHTML(replaceColorsUTF8ToHTML(original));
    }

    /**
     * Replaces all formatting and colours from a line of IRC text to UTF-8.
     *
     * @param line the input text.
     * @return the same text, but with PPF color codes formatting and colour characters.
     */
    public static String replaceCodesToUTF8(String line) {
        //todo repair this ugly fix at byte level not at charlevel
        line = line.replace('','$');
        int length = line.length();
        StringBuffer buffer = new StringBuffer();
        int i = 0;
        while (i < length) {
            char ch = line.charAt(i);
            if (ch == '$') {
                i++;
                if (i < length) {
                    ch = line.charAt(i);
                    if (Character.isDigit(ch)) {
                        buffer.append("$"+ch);
                        i++;
                    } else {
                        buffer.append('\u000F');
                    }
                } else {
                    buffer.append('\u000F');
                }
            } else {
                buffer.append(ch);
                i++;
            }
        }
        line = buffer.toString();
        line = line.replaceAll("\\$", "");
        return replaceFormattingCodesToUTF8(replaceColorsCodesToUTF8(line));
    }

    private static String replaceColorsUTF8ToHTML(String original) {
        String tmp = original;
        String key, replacement;
        for (int i = 0; i < allColorCombinations.size(); i++) {
            key = ((StringFormatRelation) allColorCombinations.get(i)).getUTF8();
            replacement = ((StringFormatRelation) allColorCombinations.get(i)).getHtml();
            int index = 0;
            key = "\\" + key;
            while (key.indexOf("$", index + 2) > -1) {
                key = key.substring(0, key.indexOf("$", index + 2)) + "\\" + key.substring(key.indexOf("$", index + 2));
                index = key.indexOf("$", index + 2);
            }
            tmp = tmp.replaceAll(key, replacement);
        }
        return tmp;
    }

    private static String replaceColorsUTF8ToCodes(String original) {
        String tmp = original;
        String key, replacement;
        for (int i = 0; i < allColorCombinations.size(); i++) {
            key = ((StringFormatRelation) allColorCombinations.get(i)).getUTF8();
            replacement = ((StringFormatRelation) allColorCombinations.get(i)).getIrcCode();
            int index = 0;
            key = "\\" + key;
            while (key.indexOf("$", index + 2) > -1) {
                key = key.substring(0, key.indexOf("$", index + 2)) + "\\" + key.substring(key.indexOf("$", index + 2));
                index = key.indexOf("$", index + 2);
            }
            tmp = tmp.replaceAll(key, replacement);
        }
        return tmp;
    }

    private static String replaceColorsCodesToUTF8(String line) {
        int length = line.length();
        StringBuffer buffer = new StringBuffer();
        int i = 0;
        while (i < length) {
            char ch = line.charAt(i);
            if (ch == '\u0003') {
                i++;
                // Skip "x" or "xy" (foreground color).
                if (i < length) {
                    ch = line.charAt(i);
                    if (Character.isDigit(ch)) {
                        String tmp = "" + ch;
                        i++;
                        if (i < length) {
                            ch = line.charAt(i);
                            if (Character.isDigit(ch)) {
                                i++;
                                if (Integer.parseInt(tmp+ch) < 16)
                                    tmp += ch;
                                else {
                                    buffer.append(COLOR_SIGNS[Integer.parseInt(tmp)]);
                                    i--;
                                    continue;
                                }
                            }
                        }
                        buffer.append(COLOR_SIGNS[Integer.parseInt(tmp)]);
                        // Now skip ",x" or ",xy" (background color).
                        if (i < length) {
                            ch = line.charAt(i);
                            if (ch == ',') {
                                i++;
                                if (i < length) {
                                    ch = line.charAt(i);
                                    if (Character.isDigit(ch)) {
                                        tmp = "" + ch;
                                        i++;
                                        if (i < length) {
                                            ch = line.charAt(i);
                                            if (Character.isDigit(ch)) {
                                                i++;
                                                if (Integer.parseInt(tmp+ch) < 16)
                                                    tmp += ch;
                                                else {
                                                    buffer.append(COLOR_SIGNS[Integer.parseInt(tmp)]);
                                                    i--;
                                                    continue;
                                                }
                                            }
                                        }
                                        buffer.append(BACKGROUND + COLOR_SIGNS[Integer.parseInt(tmp)]);
                                    } else {
                                        // Keep the comma.
                                        i--;
                                    }
                                } else {
                                    // Keep the comma.
                                    i--;
                                }
                            }
                        }
                    }
                }
            } else if (ch == '\u000F') {
                buffer.append(NORMAL);
                i++;
            } else {
                buffer.append(ch);
                i++;
            }
        }
        return buffer.toString();
    }

    private static String replaceFormattingUTF8ToHTML(String original) {
        String tmp = original;
        String key, replacement;
        for (int i = 0; i < allFormattingCombinations.size(); i++) {
            key = ((StringFormatRelation) allFormattingCombinations.get(i)).getUTF8();
            replacement = ((StringFormatRelation) allFormattingCombinations.get(i)).getHtml();
            int index = 0;
            key = "\\" + key;
            while (key.indexOf("$", index + 2) > -1) {
                key = key.substring(0, key.indexOf("$", index + 2)) + "\\" + key.substring(key.indexOf("$", index + 2));
                index = key.indexOf("$", index + 2);
            }
            tmp = tmp.replaceAll(key, replacement);
        }
        return tmp;
    }

    private static String replaceFormattingUTF8ToCode(String original) {
        String tmp = original;
        String key, replacement;
        for (int i = 0; i < allFormattingCombinations.size(); i++) {
            key = ((StringFormatRelation) allFormattingCombinations.get(i)).getUTF8();
            replacement = ((StringFormatRelation) allFormattingCombinations.get(i)).getIrcCode();
            int index = 0;
            key = "\\" + key;
            while (key.indexOf("$", index + 2) > -1) {
                key = key.substring(0, key.indexOf("$", index + 2)) + "\\" + key.substring(key.indexOf("$", index + 2));
                index = key.indexOf("$", index + 2);
            }
            tmp = tmp.replaceAll(key, replacement);
        }
        return tmp;
    }

    private static String replaceFormattingCodesToUTF8(String line) {
        int length = line.length();
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < length; i++) {
            char ch = line.charAt(i);
            if (ch == '\u000f') {
                buffer.append(NORMAL);
            } else if (ch == '\u0002') {
                buffer.append(BOLD);
            } else if (ch == '\u001f') {
                buffer.append(UNDERLINE);
            } else if (ch == '\u0016') {
                buffer.append(REVERSE);
            } else {
                buffer.append(ch);
            }
        }
        return buffer.toString();
    }

    private static String combineForeAndBackgroundToColorCode(int foregroundColorIndex, int backgroundColorIndex) {
        return COLOR_CODES[foregroundColorIndex] + "," + COLOR_CODES[backgroundColorIndex].substring(COLOR_CODES[backgroundColorIndex].length() - 2);
    }

    private static String combineForeAndBackgroundToHTMLCode(int foregroundColorIndex, int backgroundColorIndex) {
        return "<font color=\"" + COLOR_HTML[foregroundColorIndex] + "\" bgcolor=\"" + COLOR_HTML[backgroundColorIndex] + "\">";
    }

/*    Test
    public static void main(String[] args) {
        System.out.println("BOLD+CYAN+BACKGROUND+DARK_BLUE+NORMAL+DARK_GRAY+NORMAL+DARK_GREEN+GREEN");
        System.out.println(StringUtil.replaceUTF8ToCodes(BOLD+CYAN+BACKGROUND+DARK_BLUE+NORMAL+DARK_GRAY+NORMAL+DARK_GREEN+GREEN));
        System.out.println(StringUtil.replaceUTF8ToHTML(BOLD+CYAN+BACKGROUND+DARK_BLUE+NORMAL+DARK_GRAY+NORMAL+DARK_GREEN+GREEN));
        System.out.println(StringUtil.replaceCodesToUTF8("11,02140309"));
    }
*/
}
