package net.sourceforge.ppf.plugin.serverauthplugin;

// TODO: get it to wait until the server has said it has been set?  This could maybe
// stop the join (visible host) then quit and join with masked host

import net.sourceforge.ppf.PPFPlugin;
import net.sourceforge.ppf.util.PPFHelp;
import net.sourceforge.ppf.util.StringUtil;

public class ServerAuthPlugin extends PPFPlugin {
    
    private static final org.apache.commons.logging.Log LOG = org.apache.commons.logging.LogFactory
            .getLog(ServerAuthPlugin.class);

    private static final String PLUGIN_VERSION = "1.1";
    private static final String PLUGIN_ID = "SERVERAUTH";

    private String commandServerAuth = "";
    private int authLevelServerAuth = 0;
    private String authBot = "";
    private boolean messageAtStartup = false;
    private String authMessage = "";
    private boolean maskHost = false; 
    
	public ServerAuthPlugin() {
        loadDoc("ServerAuthConfig.xml");

        commandServerAuth = getTextFromDoc("/serverAuthConfig/commands/commandServerAuth");
        authLevelServerAuth = authLevelFromString(getAttributeFromDoc("/serverAuthConfig/commands/commandServerAuth", "authLevel"));

        authBot = getTextFromDoc("/serverAuthConfig/authBot");
        String msgAtStart = getAttributeFromDoc("/serverAuthConfig/authBot", "messageAtStartup");
        if (msgAtStart != null
                && (msgAtStart.equalsIgnoreCase("yes")
                        || msgAtStart.equalsIgnoreCase("true") || msgAtStart
                        .equalsIgnoreCase("on"))) {

            messageAtStartup = true;
        }
        
        authMessage = getTextFromDoc("/serverAuthConfig/authMessage");
        String strMaskHost = getTextFromDoc("/serverAuthConfig/maskHost");
        if(strMaskHost.equalsIgnoreCase("yes") || strMaskHost.equalsIgnoreCase("true") ||
            strMaskHost.equalsIgnoreCase("on")) {
            
            maskHost = true;
        }

        setVersion(PLUGIN_VERSION);

        setHelp(commandServerAuth, new PPFHelp("(Private message) "+ commandServerAuth, 
            "Auth with the network service bot", authLevelServerAuth));
	}

	public void onConnect() {
        if(messageAtStartup) {
            if (LOG.isInfoEnabled()) {
                LOG.info("[" + PLUGIN_ID + "] Authing with " + authBot);
            }
            authIt();
        }        
    }
    
    protected void onPrivateMessage(String sender, String login, String hostname,
        String message) {
            
        if(message.trim().equalsIgnoreCase(commandServerAuth)) {
            if(getBot().isAuthed(sender, authLevelServerAuth)) {
                if (LOG.isInfoEnabled()) {
                    LOG.info("[" + PLUGIN_ID + "] " + StringUtil.formatUser(sender, login, hostname)
                            + " requests authing with " + authBot);
                }
                authIt();
            }
        }
    }

    private void authIt() {
        // auth
        getBot().sendRawLine("PRIVMSG "+ authBot +" :"+ authMessage);
        // set mode
        if(maskHost) {
            getBot().sendRawLine("MODE " + getBot().getNick() + " +x");
        } else {
            getBot().sendRawLine("MODE " + getBot().getNick() + " -x");
        }
  	}
}
