// TODO: allow : in ip:port entry
// TODO: need message ERROR in server if there was an error
// TODO: with !qf, don't do !f if error with !q  
// TODO: check if the name contains a "." - otherwise assume shortcut .. then if shortcut doesn't exist, don't query but give message that SC doesn't exist

package net.sourceforge.ppf.plugin.serverqueryplugin;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import net.sourceforge.ppf.PPF;
import net.sourceforge.ppf.PPFPlugin;
import net.sourceforge.ppf.util.PPFHelp;
import net.sourceforge.ppf.util.StringUtil;
import net.sourceforge.queried.PlayerInfo;
import net.sourceforge.queried.QueriEd;
import net.sourceforge.queried.ServerInfo;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Text;
import org.jdom.xpath.XPath;

public class ServerQueryPlugin extends PPFPlugin {
    
    private static final org.apache.commons.logging.Log LOG = org.apache.commons.logging.LogFactory
            .getLog(ServerQueryPlugin.class);
    
    private static final String PLUGIN_VERSION = "2.1";
    private static final String PLUGIN_ID = "SERVERQUERY";

    private static final String SUPPORTED_GAMES = "BF, BFV, COD, D3, ET, HL, HL2, UT, UT2003, UT2004, Q3, NWN";
    
    private int queryPort = 0;
    private String commandFrags = "";
    private String commandServerQuery = "";
    private String commandServerQueryFrags = "";
    private String commandAddShortcut = "";
    private String commandRemoveShortcut = "";
    private String commandShowShortcuts = "";
    private String outputServerInfo = "";
    HashMap defaultServers = new HashMap();

    public ServerQueryPlugin() {
        Document doc = loadDoc("ServerQueryConfig.xml");
        try {
            // query port
            XPath xpPort = XPath.newInstance(
                "/serverQueryConfig/queryPort/text()");
            Text port = (Text) xpPort.selectSingleNode(doc);
            queryPort = Integer.valueOf(port.getTextTrim()).intValue();

            // command frags
            XPath xpFrags = XPath.newInstance(
                "/serverQueryConfig/commands/commandFrags/text()");
            Text cmdFrags = (Text) xpFrags.selectSingleNode(doc);
            commandFrags = cmdFrags.getTextTrim();

            // command server query
            XPath xpServerQuery = XPath.newInstance(
                "/serverQueryConfig/commands/commandServerQuery/text()");
            Text cmdServerQuery = (Text) xpServerQuery.selectSingleNode(doc);
            commandServerQuery = cmdServerQuery.getTextTrim();

            // command server query frags
            XPath xpServerQueryFrags = XPath.newInstance(
                "/serverQueryConfig/commands/commandServerQueryFrags/text()");
            Text cmdServerQueryFrags =
                (Text) xpServerQueryFrags.selectSingleNode(doc);
            commandServerQueryFrags = cmdServerQueryFrags.getTextTrim();

            // command add shortcut
            XPath xpAdd = XPath.newInstance(
                "/serverQueryConfig/commands/commandAddShortcut/text()");
            Text cmdAdd = (Text) xpAdd.selectSingleNode(doc);
            commandAddShortcut = cmdAdd.getTextTrim();

            // command remove shortcut
            XPath xpRemove = XPath.newInstance(
                "/serverQueryConfig/commands/commandRemoveShortcut/text()");
            Text cmdRemove = (Text) xpRemove.selectSingleNode(doc);
            commandRemoveShortcut = cmdRemove.getTextTrim();

            // command show shortcuts
            XPath xpShow = XPath.newInstance(
                "/serverQueryConfig/commands/commandShowShortcuts/text()");
            Text cmdShow = (Text) xpShow.selectSingleNode(doc);
            commandShowShortcuts = cmdShow.getTextTrim();

            // output - server info
            XPath xpOutputServerInfo = XPath.newInstance("/serverQueryConfig/output/serverInfo/text()");
            Text strOutputServerInfo = (Text) xpOutputServerInfo.selectSingleNode(doc);
            outputServerInfo = strOutputServerInfo.getTextTrim(); 

            // default servers
            XPath xpDefaultServers = XPath.newInstance(
                "/serverQueryConfig/defaultServers");
            List defaultServerList = xpDefaultServers.selectNodes(doc);
            Element allDefaultServers = (Element) defaultServerList.get(0);
            Iterator iter = allDefaultServers.getChildren("server").iterator();
            while (iter.hasNext()) {
                Element ds = (Element) iter.next();
                defaultServers.put(ds.getAttributeValue("name"), ds.getAttributeValue("type") +
                    " "+ ds.getAttributeValue("ip") +":"+ ds.getAttributeValue("port"));
                getItemList().put(ds.getAttributeValue("name"), ds.getAttributeValue("type") +
                    " "+ ds.getAttributeValue("ip") +":"+ ds.getAttributeValue("port"));
            }

            // shortcuts
            XPath xpShortcuts = XPath.newInstance(
                "/serverQueryConfig/shortcuts");
            List shortcutList = xpShortcuts.selectNodes(doc);
            Element allShortcuts = (Element) shortcutList.get(0);
            Iterator iter2 = allShortcuts.getChildren("shortcut").iterator();
            while (iter2.hasNext()) {
                Element sc = (Element) iter2.next();
                getItemList().put(sc.getAttributeValue("name"), sc.getAttributeValue("type") +
                    " "+ sc.getAttributeValue("ip") +":"+ sc.getAttributeValue("port"));
            }

        } catch (JDOMException e1) {
            LOG.error("Error parsing configuration", e1);
        }

        setVersion(PLUGIN_VERSION);

        setHelp(commandFrags, new PPFHelp(commandFrags,
            "Shows the players and frags for the server being queried. Supported game types: "+ SUPPORTED_GAMES,
             PPF.AUTH_NONE));
        setHelp(commandServerQuery, new PPFHelp(commandServerQuery,
            "Shows information about the server such as game type, map, and number of players.  " +
            "Supported game types: "+ SUPPORTED_GAMES, PPF.AUTH_NONE));
        setHelp(commandServerQueryFrags, new PPFHelp(commandServerQueryFrags,
            "Shows information about the server such as game type, map, and players and frags for the server being queried",
            PPF.AUTH_NONE));
        setHelp(commandAddShortcut, new PPFHelp(commandAddShortcut,
            "Adds a shortcut to a server.  This makes it easier to query a server as it is then stored by a short name rather than full ip and port",
            PPF.AUTH_TRUSTED));
        setHelp(commandRemoveShortcut, new PPFHelp(commandRemoveShortcut,
            "This removes a shortcut permanently from the list",
            PPF.AUTH_TRUSTED));
        setHelp(commandShowShortcuts, new PPFHelp(commandShowShortcuts,
            "Shows all of the shortcuts in the list using DCC Chat", PPF.AUTH_TRUSTED));
    }

    public void onMessage(String channel, String sender, String login,
        String hostname, String message) {

        if (message.startsWith(commandServerQuery + " ")) {
            String realMsg = message.substring(commandServerQuery.length()).trim();
            QueryDetails queryDetails = getQueryDetails(realMsg);
            getBot().sendMessage(channel, formatQueryDetails(queryDetails));

        } else if(message.equals(commandServerQuery)) {  // default
            String realMsg = message.substring(commandServerQuery.length()).trim();
            Iterator iter = defaultServers.values().iterator();
            while(iter.hasNext()) {
                String serverStr = (String) iter.next();
                QueryDetails queryDetails = getQueryDetails(serverStr);
                getBot().sendMessage(channel, formatQueryDetails(queryDetails));
            }
            
        } else if(message.startsWith(commandFrags + " ")) {
            String realMsg = message.substring(commandFrags.length()).trim();
            QueryDetails queryDetails = getQueryDetails(realMsg);
            getBot().sendMessage(channel, formatPlayerDetails(queryDetails));

        } else if(message.equals(commandFrags)) {  // default
            String realMsg = message.substring(commandFrags.length()).trim();
            Iterator iter = defaultServers.values().iterator();
            while(iter.hasNext()) {
                String serverStr = (String) iter.next();
                QueryDetails queryDetails = getQueryDetails(serverStr);
                getBot().sendMessage(channel, formatPlayerDetails(queryDetails));
            }

        } else if(message.startsWith(commandServerQueryFrags + " ")) {
            String realMsg = message.substring(commandServerQueryFrags.length()).trim();
            onMessage(channel, sender, login, hostname, commandServerQuery +" "+ realMsg);
            onMessage(channel, sender, login, hostname, commandFrags +" "+ realMsg);

        } else if(message.equals(commandServerQueryFrags)) {
            onMessage(channel, sender, login, hostname, commandServerQuery);
            onMessage(channel, sender, login, hostname, commandFrags);
                    
        } else if (message.startsWith(commandAddShortcut + " ")) {
            if (getBot().isAuthed(sender, PPF.AUTH_TRUSTED)) {

                String realMsg = message.substring(commandAddShortcut.length()).trim();
                String scName = realMsg.substring(0, realMsg.indexOf(" ")).trim();
                QueryDetails queryDetails = getQueryDetails(realMsg.substring(scName.length()).trim());
                
                String scType = queryDetails.getType();
                String scIP = queryDetails.getIP();
                int scPort = queryDetails.getPort();

                if(getItemList().containsKey(scName)) {
                    getBot().sendNotice(sender, "Shortcut \""+ scName +"\" already exists.  You must remove "+
                        "it first: "+ commandRemoveShortcut +" "+ scName );
                } else {
                    Element newShortcut = new Element("shortcut");
                    newShortcut.setAttribute("name", scName);
                    newShortcut.setAttribute("type", scType);
                    newShortcut.setAttribute("ip", scIP);
                    newShortcut.setAttribute("port", scPort + "");

                    if(addToItemList("shortcuts", newShortcut, scName, scType +" "+ scIP +":"+ scPort)) {
                        getBot().sendNotice(sender, "Added shortcut \"" + scName + "\"");
                    }
                }
            }
        } else if(message.startsWith(commandRemoveShortcut +" ")) {
            if (getBot().isAuthed(sender, PPF.AUTH_TRUSTED)) {
                String realMsg =  message.substring(commandRemoveShortcut.length()).trim();

                if(removeFromItemList("shortcuts", "//shortcut[@name='" + realMsg + "']", realMsg)) {
                    getBot().sendNotice(sender, "Removed trigger \""+ realMsg +"\"");
                }
            }
        }
    }

    protected void onPrivateMessage(String sender, String login, String hostname,
        String message) {

        if (message.toUpperCase().equalsIgnoreCase(commandShowShortcuts)) {
            if (getBot().isAuthed(sender, PPF.AUTH_TRUSTED)) {
                viewItemList(sender);
            }
        }
    }

    private String formatQueryDetails(QueryDetails queryDetails) {
        String outMsg = "";
        if(queryDetails == null) {
            outMsg = "Badly formed query.  Example: "+ commandServerQuery +" game-type ip:query-port  "+
            "Supported game types: "+ SUPPORTED_GAMES;
        } else {
        
            ServerInfo serverInfo = QueriEd.serverQuery(queryPort, queryDetails.getType(),
                queryDetails.getIP(), queryDetails.getPort());
                            
            if(serverInfo != null) {
                // $name :: IP: $ip:$port :: $game :: $map :: Players: $players/$maxPlayers
                String str = outputServerInfo.replaceAll("\\$name", serverInfo.getName());
                str = str.replaceAll("\\$ip", serverInfo.getIp());
                str = str.replaceAll("\\$port", serverInfo.getPort());
                str = str.replaceAll("\\$game", serverInfo.getGame());
                str = str.replaceAll("\\$map", serverInfo.getMap());
                str = str.replaceAll("\\$players", serverInfo.getPlayerCount());
                str = str.replaceAll("\\$maxplayers", serverInfo.getMaxPlayers());
                                
                outMsg = StringUtil.replaceUTF8ToCodes(str);
            } else {
                outMsg = "Error querying server ["+ queryDetails.getType() +"]: "+ 
                    queryDetails.getIP() +":"+ queryDetails.getPort();
            }
        }
        
        return outMsg;
    }
 
    private String formatPlayerDetails(QueryDetails queryDetails) {
        StringBuffer msg = new StringBuffer("");
        if(queryDetails == null) {
            msg.append("Badly formed query.  Example: "+ commandFrags +" game-type ip:query-port" +
            "  Supported game types: "+ SUPPORTED_GAMES);
        } else {
        
            ArrayList playerInfo = QueriEd.playerQuery(queryPort, queryDetails.getType(),
                queryDetails.getIP(), queryDetails.getPort());
                        
            if(playerInfo != null) {
                Iterator iter = playerInfo.iterator();
                int count = 0;
                while(iter.hasNext()) {
                    PlayerInfo pInfo = (PlayerInfo) iter.next();
                    if(count > 0) {
                        msg.append(", ");
                    }
                    count++;
                    msg.append(pInfo.getName() +" [");
                    if(pInfo.getScore() != -9999) {
                        msg.append(pInfo.getScore());
                    }
                    if(pInfo.getKills() != -9999) {
                        if(pInfo.getScore() != -9999) {
                            msg.append("/");
                        }
                        msg.append(pInfo.getKills());
                    }
                    if(pInfo.getDeaths() != -9999) {
                        if(pInfo.getDeaths() != -9999) {
                            msg.append("/");
                        }
                        msg.append(pInfo.getDeaths());
                    }
                    msg.append("]");
                }
                if(count == 0) {
                    msg.append("No players");
                }
            } else {
                msg.append("Error querying server ["+ queryDetails.getType() +"]: "+ 
                    queryDetails.getIP() +":"+ queryDetails.getPort());
            }
        }
        
        return msg.toString();
    }
       
    private QueryDetails getQueryDetails(String msg) {
        String type = "HL";
        String ip = "";
        int port = 27015;

        if (msg.length() > 0) {
            if (getItemList().containsKey(msg)) {
                msg = (String) getItemList().get(msg);
            } else if(defaultServers.containsKey(msg)) {
                msg = (String) defaultServers.get(msg);
            }

            int typeIndex = msg.indexOf(" ");
            if(typeIndex > 0) {
                type = msg.substring(0, typeIndex).trim();
            } else {
                // not formed correctly
                return null;
            }
            
            msg = msg.substring(type.length()).trim();
            
            int portIndex = msg.indexOf(":");
            if (portIndex > 0) {
                try {
                    ip = msg.substring(0, portIndex);
                    port = Integer.valueOf(msg.substring(msg.indexOf(":") + 1).trim()).intValue();
                } catch (NumberFormatException nfex) {
                    // could get a bad port ... so set default
                    port = 27015;
                }
            } else { // maybe a space between ip and port
                int spaceIndex = msg.indexOf(" ");
                if(spaceIndex > 0) {
                    try {
                        ip = msg.substring(0, spaceIndex).trim();
                        port = Integer.valueOf(msg.substring(spaceIndex).trim()).intValue();
                    } catch (NumberFormatException nfex) {
                        // could get a bad port ... so set default
                        port = 27015;                        
                    }
                } else {
                    ip = msg;
                }
            }       
        }

        return new QueryDetails(type, ip, port);
    }
    
    private static class QueryDetails {
        private String type;
        private String ip;
        private int port;
        
        public QueryDetails(String type, String ip, int port) {
            this.type = type;
            this.ip = ip;
            this.port = port;
        }

        public String getType() {
            return type;
        }
                 
        public String getIP() {
            return ip;
        }
        
        public int getPort() {
            return port;
        }
    }
}